// MOLDYN.H : molecular dynamics classes.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef MOLDYN_H
#define MOLDYN_H

class moldyn;
class moldyn_param;

/*################################################################################################*/

#include "engine.h"

/*################################################################################################*/

/**	A molecular dynamics class.

	So far very long simulations have not been tested, so possile translation/rotation 
	problems are not solved... current solution is to start simulations from 0 K -> 
	no translation/rotation in the initial state -> conservation of linear and angular 
	momentum -> no need to worry at all... but for how long it will work???
*/

class moldyn
{
	protected:
	
	engine * eng;
	
	f64 * vel;
	f64 * acc;
	
	f64 * mass;
	
	f64 temp;
	
	f64 step1; f64 step2;
	f64 kin; f64 pot;
	
	friend class model;
	
	public:
	
	moldyn(engine *, f64, f64);
	~moldyn(void);
	
	/**	This is a "velocity-Verlet"-type integrator...
	
		Allen MP, Tildesley DJ : "##Computer Simulation of Liquids", Clarendon Press, 1987
	*/
	
	void TakeMDStep(void);
	
	f64 KineticEnergy(void);
	
	f64 ConvTempEKin(f64);
	f64 ConvEKinTemp(f64);
	
	void SetEKin(f64);
};

class moldyn_param
{
//	protected:
	public:		// in TARGET3 we have to access this from a function -> gets complicated. fix later...
	
	bool confirm;
	
	friend class model;
	
	public:
	
	i32s nsteps;
	f64 temperature;
	f64 timestep;
	
	char filename[256];
	
	public:
	
	moldyn_param(void) { confirm = false; }
	~moldyn_param(void) { }
};

/*################################################################################################*/

#endif	// MOLDYN_H

// eof
