// TAB_MM_DEFAULT.H : tables for the molecular mechanics force field.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef TAB_MM_DEFAULT_H
#define TAB_MM_DEFAULT_H

class default_tables;

/*################################################################################################*/

class model;		// model.h
class typerule;		// typerule.h

#include "eng1_mm_default.h"

#include <vector>
using namespace std;

/*################################################################################################*/

struct default_at		// atomtype
{
	i32s atomtype;
	typerule * tr;
	
	char * description;
};

struct default_bs		// bond stretching
{
	i32s atmtp[2];
	bondtype bndtp;
	f64 param[2];
};

struct default_ab		// angle bending
{
	i32s atmtp[3];
	bondtype bndtp[2];
	f64 param[2];
};

struct default_tr		// torsion
{
	i32s atmtp[4];
	bondtype bndtp[3];
	f64 k; f64 s;
};

// out of plane...

struct default_lj		// lennard-jones
{
	i32s type;
	f64 r; f64 k;
};

struct default_ci		// charge increment
{
	i32s atmtp[2];
	bondtype bndtp;
	f64 delta;
};

/*################################################################################################*/

/// Molecular mechanics force field parameter tables.

class default_tables
{
	private:
	
	static default_tables * instance;
	
	vector<default_at> at_vector;
	
	vector<default_bs> bs_vector;
	vector<default_ab> ab_vector;
	vector<default_tr> tr_vector;
	vector<default_lj> lj_vector;
	vector<default_ci> ci_vector;
	
	private:
	
	default_tables(void);
	
	public:
	
	~default_tables(void);
	static default_tables * GetInstance(void);
	
	f64 GetChargeInc(bond *, ostream *);
	
	bool Init(eng1_mm *, mm_default_bt1 *, i32s);
	bool Init(eng1_mm *, mm_default_bt2 *, i32s *);
	bool Init(eng1_mm *, mm_default_bt3 *, i32s *);
	
	bool Init(eng1_mm *, mm_default_nbt1 *, bool);
	
	i32s UpdateTypes(model *);
	void PrintAllTypeRules(ostream &);
};

/*################################################################################################*/

#endif	// TAB_MM_DEFAULT_H

// eof
