// PROJECT.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "project.h"	// config.h is here -> we get ENABLE-macros here...

#ifdef ENABLE_GRAPHICS
#include "views.h"

#include "v3d.h"

#include "atom.h"
#include "bond.h"

#include "engine.h"
#include "geomopt.h"
#include "intcrd.h"

#include "utility.h"

#include "plane.h"
#include "surface.h"

#include "color.h"
#include "views.h"

#include "Vector.h"	// OpenBabel utils for torsion, angle measure
using namespace OpenBabel;

#include <iomanip>
#include <algorithm>
#include <fstream>
#include <strstream>
using namespace std;

/*################################################################################################*/

iGLu project::list_counter = 1;		// zero is not a valid display list number...

color_mode_element project::cm_element = color_mode_element();
color_mode_state project::cm_state = color_mode_state();

project::project(graphics_class_factory & p1) : model(p1)
{
	graphics_factory = & p1;
	
	project_path = NULL;
	project_filename = NULL;
	
	SetDefaultProjectFileName();
	
	select_buffer = new iGLu[SB_SIZE];
	selected_object = NULL;
	
#ifdef ENABLE_TREELIST_VIEW

	tlv = NULL;
	
#endif	// ENABLE_TREELIST_VIEW

}

project::~project(void)
{
	selected_object = NULL;
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) DestroyTreeListView();
	
#endif	// ENABLE_TREELIST_VIEW

	while (graphics_view_vector.size() > 0)
	{
		graphics_view * ref;
		ref = graphics_view_vector.back();
		RemoveGraphicsView(ref, true);
	}
	
	while (object_vector.size() > 0)
	{
		smart_object * ref;
		ref = object_vector.back();
		RemoveObject(ref);
	}
	
	if (project_path != NULL) delete[] project_path;
	if (project_filename != NULL) delete[] project_filename;
	
	delete[] select_buffer;
}

bool project::AddTP(void * owner, transparent_primitive & tp)
{
	if (!tp.TestOwner(owner)) return false;		// this is just a double check, to make
	tp_vector.push_back(tp); return true;		// sure that we have correct "owner"...
}

void project::UpdateMPsForAllTPs(void * owner)
{
	for (i32u n1 = 0;n1 < tp_vector.size();n1++)
	{
		if (tp_vector[n1].TestOwner(owner)) tp_vector[n1].GetData()->UpdateMP();
	}
}

void project::RemoveAllTPs(void * owner)
{
	i32u n1 = 0;
	while (n1 < tp_vector.size())
	{
		vector<transparent_primitive>::iterator iter;
		
		if (!tp_vector[n1].TestOwner(owner))
		{
			n1++;
		}
		else
		{
			delete tp_vector[n1].GetData();		// destroy the data object!!!
			
			iter = (tp_vector.begin() + n1);
			tp_vector.erase(iter);
		}
	}
}

void project::SetProjectPath(const char * path)
{
	if (project_path != NULL) delete[] project_path;
	
	project_path = new char[strlen(path) + 1];
	strcpy(project_path, path);
}

void project::SetProjectFileName(const char * filename)
{
	if (project_filename != NULL) delete[] project_filename;
	
	project_filename = new char[strlen(filename) + 1];
	strcpy(project_filename, filename);
}

void project::SetDefaultProjectFileName(void)
{
	static i32s id_counter = 1;
	
	char buffer[1024];
	ostrstream str(buffer, sizeof(buffer));
	str << "untitled" << setw(2) << setfill('0') << id_counter++ << ends;
	
	SetProjectFileName(buffer);
}

void project::ParseProjectFileNameAndPath(const char * string)
{
	char * localstring1 = new char[strlen(string) + 1];
	strcpy(localstring1, string);
	
	i32s lastdir = NOT_DEFINED;
	for (i32s n1 = 0;n1 < (i32s) strlen(localstring1);n1++)
	{
		if (localstring1[n1] == DIR_SEPARATOR) lastdir = n1;
	}
	
	char * localstring2 = localstring1;
	
	// set project_path if needed...
	// set project_path if needed...
	// set project_path if needed...
	
	if (lastdir != NOT_DEFINED)
	{
		localstring2[lastdir] = 0;
		SetProjectPath(localstring2);
		
		localstring2 = & localstring2[lastdir + 1];
	}
	
	// and now set project_filename, without extension...
	// and now set project_filename, without extension...
	// and now set project_filename, without extension...
	
	i32s lastext = NOT_DEFINED;
	for (i32s n1 = 0;n1 < (i32s) strlen(localstring2);n1++)
	{
		if (localstring2[n1] == EXT_SEPARATOR) lastext = n1;
	}
	
	if (lastext != NOT_DEFINED)
	{
	  // This only removes an extension if it matches *our* extension,
	  // which makes problems for imported files e.g. nh3.mol.mmg1p (!) instead of nh3.mol or nh3.mm1gp
	  //		char * localstring3 = & localstring2[lastext + 1];
	  //		bool has_extension = !strcmp(localstring3, GetProjectFileNameExtension());
	  //		if (has_extension) localstring2[lastext] = 0;

	  // instead
	  localstring2[lastext] = 0;
	}
	
	SetProjectFileName(localstring2);
	
	delete[] localstring1;
}

void project::GetProjectFileName(char * buffer, int buffer_size, bool with_extension)
{
	ostrstream ostr(buffer, buffer_size);
	
	ostr << project_filename;
	if (with_extension) ostr << EXT_SEPARATOR << GetProjectFileNameExtension();
	ostr << ends;
}

void project::GetFullProjectFileName(char * buffer, int buffer_size)
{
	ostrstream ostr(buffer, buffer_size);
	
	if (project_path != NULL) ostr << project_path << DIR_SEPARATOR;
	ostr << project_filename << EXT_SEPARATOR << GetProjectFileNameExtension() << ends;
}

iGLu project::GetDisplayListIDs(iGLu p1)
{
	iGLu first = list_counter;
	list_counter += p1;
	
	return first;
}

void project::DeleteDisplayLists(iGLu p1, iGLu p2)
{
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		graphics_view_vector[n1]->SetCurrent();
		glDeleteLists(p1, p2);
	}
}

/*##############################################*/
/*##############################################*/

i32s project::IsLight(const dummy_object * p1)
{
	i32s index = NOT_DEFINED;
	for (i32u n1 = 0;n1 < light_vector.size();n1++)
	{
		if (light_vector[n1] == p1) index = n1;
	}
	
	return index;
}

bool project::SelectLight(const dummy_object * p1)
{
	i32s n1 = IsLight(p1);
	if (n1 < 0) return false;
	
	selected_object = light_vector[n1];
	event_SelectedObjectChanged();
	
#ifdef TARGET2

	cout << "selected " << light_vector[n1]->GetObjectName() << " which is ";
	cout << (light_vector[n1]->owner != NULL ? "local " : "global ") << endl;
	
#endif	// TARGET2

	return true;
}

bool project::AddGlobalLight(light * p1)
{
	iGLs max_local_size = 0;
	for (i32u n1 = 0;n1 < camera_vector.size();n1++)
	{
		iGLs local_size = CountLocalLights(camera_vector[n1]);
		if (local_size > max_local_size) max_local_size = local_size;
	}
	
	iGLs total_lights = CountGlobalLights() + max_local_size;
	iGLs max_lights; glGetIntegerv(GL_MAX_LIGHTS, & max_lights);
	if (total_lights == max_lights) return false;
	
	light_vector.push_back(p1);
	SetGlobalLightNumbers();
	
	for (i32u n1 = 0;n1 < camera_vector.size();n1++)
	{
		SetLocalLightNumbers(camera_vector[n1]);
		SetupLights(camera_vector[n1]);
	}
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) tlv->LightAdded(p1);
	
#endif	// ENABLE_TREELIST_VIEW

	selected_object = light_vector.back();
	event_SelectedObjectChanged();
	
	return true;
}

bool project::AddLocalLight(light * p1, camera * p2)
{
	iGLs total_lights = CountGlobalLights() + CountLocalLights(p2);
	iGLs max_lights; glGetIntegerv(GL_MAX_LIGHTS, & max_lights);
	if (total_lights == max_lights) return false;
	
	p1->owner = p2; light_vector.push_back(p1);
	SetLocalLightNumbers(p2); SetupLights(p2);
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) tlv->LightAdded(p1);
	
#endif	// ENABLE_TREELIST_VIEW

	selected_object = light_vector.back();
	event_SelectedObjectChanged();
	
	return true;
}

bool project::RemoveLight(dummy_object * p1)
{
	i32s n1 = IsLight(p1);
	if (n1 < 0) return false;
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) tlv->LightRemoved(light_vector[n1]);
	
#endif	// ENABLE_TREELIST_VIEW

	camera * owner = light_vector[n1]->owner;
	light_vector.erase(light_vector.begin() + n1);
	delete p1;
	
	if (owner != NULL)
	{
		SetLocalLightNumbers(owner);
		SetupLights(owner);
	}
	else
	{
		SetGlobalLightNumbers();
		for (i32u n1 = 0;n1 < camera_vector.size();n1++)
		{
			SetLocalLightNumbers(camera_vector[n1]);
			SetupLights(camera_vector[n1]);
		}
	}
	
	return true;
}

iGLs project::CountGlobalLights(void)
{
	iGLs sum = 0; i32u n1 = 0;
	while (n1 < light_vector.size())
	{
		if (light_vector[n1++]->owner == NULL) sum++;
	}
	
	return sum;
}

iGLs project::CountLocalLights(camera * p1)
{
	iGLs sum = 0; i32u n1 = 0;
	while (n1 < light_vector.size())
	{
		if (light_vector[n1++]->owner == p1) sum++;
	}
	
	return sum;
}

bool project::SetupLights(camera * p1)
{
	if (p1->prj != this) return false;
	iGLs max_lights; glGetIntegerv(GL_MAX_LIGHTS, & max_lights);
	
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		if (graphics_view_vector[n1]->cam != p1) continue;
		
		graphics_view_vector[n1]->SetCurrent();
		for (iGLs n2 = 0;n2 < max_lights;n2++)
		{
			glDisable((GLenum) (GL_LIGHT0 + n2));
		}
		
		for (i32u n2 = 0;n2 < light_vector.size();n2++)
		{
			bool test1 = (light_vector[n2]->owner != NULL);
			bool test2 = (light_vector[n2]->owner != p1);
			if (test1 && test2) continue;
			
			light_vector[n2]->SetupProperties(); bool test = false;
			if (light_vector[n2]->owner == NULL && p1->use_global_lights) test = true;
			if (light_vector[n2]->owner == p1 && p1->use_local_lights) test = true;
			if (test) glEnable((GLenum) light_vector[n2]->number);
		}
	}
	
	return true;
}

void project::SetGlobalLightNumbers(void)
{
	iGLs counter = 0;
	for (i32u n1 = 0;n1 < light_vector.size();n1++)
	{
		if (light_vector[n1]->owner != NULL) continue;
		light_vector[n1]->number = GL_LIGHT0 + counter++;
	}
}

void project::SetLocalLightNumbers(camera * p1)
{
	iGLs counter = CountGlobalLights();
	for (i32u n1 = 0;n1 < light_vector.size();n1++)
	{
		if (light_vector[n1]->owner != p1) continue;
		light_vector[n1]->number = GL_LIGHT0 + counter++;
	}
}

/*##############################################*/
/*##############################################*/

i32s project::IsObject(const dummy_object * p1)
{
	i32s index = NOT_DEFINED;
	for (i32u n1 = 0;n1 < object_vector.size();n1++)
	{
		if (object_vector[n1] == p1) index = n1;
	}
	
	return index;
}

bool project::SelectObject(const dummy_object * p1)
{
	i32s n1 = IsObject(p1);
	if (n1 < 0) return false;
	
	selected_object = object_vector[n1];
	event_SelectedObjectChanged();
	
#ifdef TARGET2

	cout << "selected object #" << n1 << " which is ";
	cout << object_vector[n1]->GetObjectName() << endl;
	
#endif	// TARGET2

	return true;
}

void project::AddObject(smart_object * p1)
{
	object_vector.push_back(p1);
	selected_object = object_vector.back();
	event_SelectedObjectChanged();
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) tlv->ObjectAdded(p1);
	
#endif	// ENABLE_TREELIST_VIEW

}

bool project::RemoveObject(dummy_object * p1)
{
	i32s n1 = IsObject(p1);
	if (n1 < 0) return false;
	
#ifdef ENABLE_TREELIST_VIEW

	if (tlv != NULL) tlv->ObjectRemoved(object_vector[n1]);
	
#endif	// ENABLE_TREELIST_VIEW

	object_vector.erase(object_vector.begin() + n1);
	delete p1; return true;
}

// these are the measuring functions, that only take coordinates as input (so they are model-independent)...
// these are the measuring functions, that only take coordinates as input (so they are model-independent)...
// these are the measuring functions, that only take coordinates as input (so they are model-independent)...

float measure_len(float * c1, float * c2)
{
	v3d<float> v1(c1, c2);
	return v1.len();
}

float measure_ang(float * c1, float * c2, float * c3)
{
	Vector v1, v2;
	v1 = Vector(c1[0] - c2[0], c1[1] - c2[1], c1[2] - c2[2]);
	v2 = Vector(c3[0] - c2[0], c3[1] - c2[1], c3[2] - c2[2]);
	return VectorAngle(v1, v2);
}

float measure_tor(float * c1, float * c2, float * c3, float * c4)
{
	Vector v1, v2, v3, v4;
	v1 = Vector(c1[0], c1[1], c1[2]) * 10.0f;
	v2 = Vector(c2[0], c2[1], c2[2]) * 10.0f;
	v3 = Vector(c3[0], c3[1], c3[2]) * 10.0f;
	v4 = Vector(c4[0], c4[1], c4[2]) * 10.0f;
	return CalcTorsionAngle(v1, v2, v3, v4);
}

/*##############################################*/
/*##############################################*/

graphics_view * project::AddGraphicsView(bool detached)
{
	fGL focus = GetDefaultFocus();
	camera_vector.push_back(new camera(ol_static(), focus, this));
	
	graphics_view_vector.push_back(GetGraphicsFactory()->ProduceGraphicsView(this, camera_vector.back(), detached));
	
	// also add a default light, so that we can see something...
	// also add a default light, so that we can see something...
	// also add a default light, so that we can see something...
	
	AddLocalLight(new directional_light(ol_static()), camera_vector.back());
	
	UpdateAllWindowTitles();
	return graphics_view_vector.back();
}

// here the camera pointer "p1" should already be in the camera_vector; see the above case!!!

graphics_view * project::AddGraphicsView(camera * p1, bool detached)
{
	graphics_view_vector.push_back(GetGraphicsFactory()->ProduceGraphicsView(this, p1, detached));
	
	// setup the old lights that we already have...
	// setup the old lights that we already have...
	// setup the old lights that we already have...
	
	SetupLights(p1);
	
	UpdateAllWindowTitles();
	return graphics_view_vector.back();
}

bool project::IsThisLastGraphicsView(graphics_view * p1)
{
	i32s other_views_for_this_cam = 0;
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		bool test1 = (graphics_view_vector[n1] == p1);
		bool test2 = (graphics_view_vector[n1]->cam == p1->cam);
		
		if (!test1 && test2) other_views_for_this_cam++;
	}
	
	if (!other_views_for_this_cam && camera_vector.size() < 2) return true;
	else return false;
}

bool project::RemoveGraphicsView(graphics_view * p1, bool force)
{
	i32s n1 = NOT_DEFINED;
	for (i32u n2 = 0;n2 < camera_vector.size();n2++)
	{
		if (camera_vector[n2] == p1->cam) n1 = n2;
	}
	
	if (n1 == NOT_DEFINED)		// this should never happen...
	{
		cout << "WARNING: could not find the camera at project::RemoveGraphicsView()." << endl;
		return false;
	}
	
	i32s other_views_for_this_cam = 0;	// this is also calculated at IsThisLastGraphicsView().
	
	i32s n2 = NOT_DEFINED;
	for (i32u n3 = 0;n3 < graphics_view_vector.size();n3++)
	{
		if (graphics_view_vector[n3] == p1) n2 = n3;
		else if (graphics_view_vector[n3]->cam == p1->cam) other_views_for_this_cam++;
	}
	
	if (n2 == NOT_DEFINED)		// this should never happen...
	{
		cout << "WARNING: could not find the view at project::RemoveGraphicsView()." << endl;
		return false;
	}
	
	if (!force && (!other_views_for_this_cam && camera_vector.size() < 2))	// refuse to close the last window!!!
	{
		err->ErrorMessage("This is the last graphics view for\nthis project - can't close it.");
		return false;
	}
	
	// if this was the last view for a camera, we have to remove also the camera.
	// but before that we must remove all local lights that relate to the camera...
	
	// a light object can also be selected_object, so we must compare the pointers
	// and invalidate selected_object if necessary...
	
	if (!other_views_for_this_cam)
	{
		i32u n3 = 0;
		while (n3 < light_vector.size())
		{
			if (light_vector[n3]->owner == camera_vector[n1])
			{
				if (selected_object != NULL)
				{
					if (selected_object == light_vector[n3])
					{
						selected_object = NULL;
						event_SelectedObjectChanged();
					}
				}
				
				#ifdef ENABLE_TREELIST_VIEW
				
				if (tlv != NULL) tlv->LightRemoved(light_vector[n3]);
				
				#endif	// ENABLE_TREELIST_VIEW
				
				delete light_vector[n3];
				light_vector.erase(light_vector.begin() + n3);
			}
			else n3++;
		}
		
		SetGlobalLightNumbers();
		for (n3 = 0;n3 < camera_vector.size();n3++)
		{
			if (n1 == (i32s) n3) continue;
			
			SetLocalLightNumbers(camera_vector[n3]);
			SetupLights(camera_vector[n3]);
		}
	}
	
	// now it's time to remove the window...
	
	delete graphics_view_vector[n2];
	graphics_view_vector.erase(graphics_view_vector.begin() + n2);
	
	// and now we finally can remove the camera, if needed...
	
	if (!other_views_for_this_cam)
	{
		delete camera_vector[n1];
		camera_vector.erase(camera_vector.begin() + n1);
	}
	
	UpdateAllWindowTitles();
	return true;
}

#ifdef ENABLE_TREELIST_VIEW

treelist_view * project::CreateTreeListView(void)
{
	if (tlv != NULL) return tlv;
	else
	{
		tlv = GetGraphicsFactory()->ProduceTreeListView(this);
		return tlv;
	}
}

bool project::DestroyTreeListView(void)
{
	if (!tlv) return false;
	else
	{
		delete tlv;
		tlv = NULL;
		
		return true;
	}
}

#endif	// ENABLE_TREELIST_VIEW

enlevdiag_view * project::AddEnLevDiagView(void)
{
	enlevdiag_view * eldv = graphics_factory->ProduceEnLevDiagView(this);
	plotting_view_vector.push_back(eldv);
	
	return eldv;
}

void project::UpdateAllViews(void)
{
	UpdateAllGraphicsViews();
}

void project::UpdateAllWindowTitles(void)
{
        char buffer[1024];

	for (i32u n1 = 0;n1 < camera_vector.size();n1++)
	{
		i32s count = 1;
		for (i32u n2 = 0;n2 < graphics_view_vector.size();n2++)
		{
			if (graphics_view_vector[n2]->cam != camera_vector[n1]) continue;
			
			ostrstream str(buffer, sizeof(buffer));
			str << "camera #" << (n1 + 1) << " view #" << count++ << ends;
			graphics_view_vector[n2]->SetTitle(buffer);
			
			// should we also update treelist view title, if it exists? or does it have a fixed name???
			// should we also update treelist view title, if it exists? or does it have a fixed name???
			// should we also update treelist view title, if it exists? or does it have a fixed name???
		}
	}
}

void project::UpdateAllGraphicsViews(bool flag)
{
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		graphics_view_vector[n1]->Update(flag);
	}
}

void project::UpdateGraphicsViews(camera * cam, bool flag)
{
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		if (graphics_view_vector[n1]->cam != cam) continue;
		graphics_view_vector[n1]->Update(flag);
	}
}

void project::UpdateGraphicsView(graphics_view * gv, bool flag)
{
	gv->Update(flag);
}

void project::UpdatePlottingViews(void)
{
	for (i32u n1 = 0;n1 < plotting_view_vector.size();n1++)
	{
		// energy-level diagrams might require scaling here?!?!?!
		// detect them and call the scaling function... ANY BETTER WAY TO DO THIS???
		enlevdiag_view * eldv = dynamic_cast<enlevdiag_view *>(plotting_view_vector[n1]);
		if (eldv != NULL) eldv->SetCenterAndScale();
		
		plotting_view_vector[n1]->Update();
	}
}

void project::RenderAllTPs(graphics_view * gv, rmode)
{
	// here we will render those transparent primitives...
	// all models should call this before returning from their versions!!!
	
	// first we must update the distances for all TP's, and sort them...
	
	const fGL * ref1 = gv->cam->GetLocData()->crd;
	const fGL * ref2 = gv->cam->GetLocData()->zdir.data;
	
	for (i32u n1 = 0;n1 < tp_vector.size();n1++)
	{
		tp_vector[n1].UpdateDistance(ref1, ref2);
	}
	
	sort(tp_vector.begin(), tp_vector.end());
	
	// and then we will just render them...
	// it looks better if we disable depth buffer changes...
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask(false); glEnable(GL_BLEND);
	
	for (i32u n1 = 0;n1 < tp_vector.size();n1++)
	{
		tp_vector[n1].GetData()->Render();
	}
	
	glDisable(GL_BLEND); glDepthMask(true);
}

/*##############################################*/
/*##############################################*/

void project::ProcessCommandString(graphics_view * gv, const char * command)
{
	char mbuff1[512];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	str1 << "Processing Command : " << command << endl << ends;
	PrintToLog(mbuff1);
	
	istrstream istr(command);
	char kw1[32]; istr >> kw1;	// the 1st keyword.
	
	if (!strcmp("help", kw1))
	{
		PrintToLog("> AVAILABLE COMMANDS:\n");		// use alphabetical order???
		
		PrintToLog("> add light (local/global) (directional/spotlight) -- add a new light object.\n");
		
		PrintToLog("> add plane <vf> <cf> <cscale> <dim> <res> <tp> <alpha> -- add a plane object.\n");
		PrintToLog(">   where: <vf> = value function : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <cf> = colour function : red green blue rb1 rb2\n");
		PrintToLog(">          <cscale> = scaling value for calculating the colours\n");
		PrintToLog(">          <dim> = dimension of the plane object (in nm units)\n");
		PrintToLog(">          <res> = resolution of the plane object\n");
		PrintToLog(">          <tp> = 0 or 1 telling if the object is transparent\n");
		PrintToLog(">          <alpha> = transparency alpha value\n");
		
		PrintToLog("> add volrend <vf> <cf> <cscale> <dim> <res> <alpha> -- add a volume-rendering object.\n");
		PrintToLog(">   where: <vf> = value function : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <cf> = colour function : red green blue rb1 rb2\n");
		PrintToLog(">          <cscale> = scaling value for calculating the colours\n");
		PrintToLog(">          <dim> = dimension of the plane object (in nm units)\n");
		PrintToLog(">          <res> = resolution of the plane object\n");
		PrintToLog(">          <alpha> = transparency alpha value\n");
		
		PrintToLog("> add surf1 <vf1> <vf2> <cf> <sscale> <cscale> <dim> <res> <solid> <tp> <alpha> -- add a single surface object.\n");
		PrintToLog(">   where: <vf1> = value function for calculating the surface : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <vf2> = value function for calculating the colours : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <cf> = colour function : red green blue rb1 rb2\n");
		PrintToLog(">          <sscale> = scaling value for calculating the surface\n");
		PrintToLog(">          <cscale> = scaling value for calculating the colours\n");
		PrintToLog(">          <dim> = dimension of the plane object (in nm units)\n");
		PrintToLog(">          <res> = resolution of the plane object\n");
		PrintToLog(">          <solid> = 0 or 1 telling if the object is solid\n");
		PrintToLog(">          <tp> = 0 or 1 telling if the object is transparent\n");
		PrintToLog(">          <alpha> = transparency alpha value\n");
		
		PrintToLog("> add surf2 <vf1> <vf2> <cf1> <cf2> <sscale1> <sscale2> <cscale> <dim> <res> <solid> <tp> <alpha> -- add a pair of surface objects.\n");
		PrintToLog(">   where: <vf1> = value function for calculating the surface : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <vf2> = value function for calculating the colours : esp vdws eldens mo mod unity\n");
		PrintToLog(">          <cf1> = colour function for 1st surface : red green blue rb1 rb2\n");
		PrintToLog(">          <cf2> = colour function for 2nd surface : red green blue rb1 rb2\n");
		PrintToLog(">          <sscale1> = scaling value for calculating the surface for 1st surface\n");
		PrintToLog(">          <sscale2> = scaling value for calculating the surface for 2nd surface\n");
		PrintToLog(">          <cscale> = scaling value for calculating the colours\n");
		PrintToLog(">          <dim> = dimension of the plane object (in nm units)\n");
		PrintToLog(">          <res> = resolution of the plane object\n");
		PrintToLog(">          <solid> = 0 or 1 telling if the object is solid\n");
		PrintToLog(">          <tp> = 0 or 1 telling if the object is transparent\n");
		PrintToLog(">          <alpha> = transparency alpha value\n");
		
		PrintToLog("> help -- print all available commands in command strings.\n");
		
		PrintToLog("> random_search <cycles> <optsteps> -- perform a random conformational search.\n");
		PrintToLog("> systematic_search <divisions> <optsteps> -- perform a systematic conformational search.\n");
		PrintToLog("> montecarlo_search <init_cycles> <simul_cycles> <optsteps> -- perform a MonteCarlo search.\n");
		
		PrintToLog("> make_plot1 A B C D <div> <start_ang> <end_ang> <optsteps> -- create a 1D energy vs. torsion plot.\n");
		PrintToLog("> make_plot2 A B C D <div> <start_ang> <end_ang> I J K L <div> <start_ang> <end_ang> <optsteps> -- create a 2D energy vs. torsions plot.\n");
		
		PrintToLog("> set_current_orbital <orbital_index> -- set the current orbtal index for plotting the orbitals.\n");

		PrintToLog("> build_amino <sequence> (helix/strand) -- amino acid sequence builder.\n");
		PrintToLog("> build_nucleic <sequence> -- nucleic acid sequence builder.\n");
		
		return;
	}
	
	if (!strcmp("add", kw1))
	{
		char kw2[32]; istr >> kw2;	// the 2nd keyword; type of the object to add.
		char object_name[128];
		
		if (!strcmp("light", kw2))
		{
			char kw3[32]; istr >> kw3;
			char kw4[32]; istr >> kw4;
			
			bool is_local = true;
			if (kw3[0] == 'g') is_local = false;
			
			bool is_directional = true;
			if (kw4[0] == 's') is_directional = false;
			
			light * new_light;
			if (is_directional) new_light = new directional_light(ol_static());
			else
			{
				new_light = new spot_light(ol_static());
				
				const fGL trans[3] = { 0.0, 0.0, -1.0 };
				new_light->TranslateObject((const fGL *) trans, gv->cam->GetLocData());
			}
			
			if (!is_local) AddGlobalLight(new_light);
			else AddLocalLight(new_light, gv->cam);
			
			if (!is_local || graphics_view::draw_info) UpdateAllGraphicsViews();
			else if (is_local) UpdateGraphicsViews(gv->cam);
			
			ostrstream strR(mbuff1, sizeof(mbuff1));
			strR << "Added a new object : light (";
			strR << (is_local ? "local" : "global") << " ";
			strR << (is_directional ? "directional" : "spotlight") << ")." << endl << ends;
			PrintToLog(mbuff1);
			return;
		}
		
		if (!strcmp("plane", kw2))
		{
			char kw3[32]; istr >> kw3;
			char kw4[32]; istr >> kw4;
			char kw5[32]; istr >> kw5;
			char kw6[32]; istr >> kw6;
			char kw7[32]; istr >> kw7;
			char kw8[32]; istr >> kw8;
			char kw9[32]; istr >> kw9;
			char ** endptr = NULL;
			
			cp_param cpp;
			cpp.prj = this; cpp.ref = GetCurrentSetup()->GetCurrentEngine();
			
			if (!strcmp(kw3, "esp")) cpp.vf = (ValueFunction *) value_ESP;
			else if (!strcmp(kw3, "vdws")) cpp.vf = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw3, "eldens")) cpp.vf = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw3, "mo")) cpp.vf = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw3, "mod")) cpp.vf = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw3, "unity")) cpp.vf = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add plane : unknown value function.\n"); return; }
			
			if (!strcmp(kw4, "red")) cpp.cf = (ColorFunction *) GetRedColor;
			else if (!strcmp(kw4, "green")) cpp.cf = (ColorFunction *) GetGreenColor;
			else if (!strcmp(kw4, "blue")) cpp.cf = (ColorFunction *) GetBlueColor;
			else if (!strcmp(kw4, "rb1")) cpp.cf = (ColorFunction *) GetRBRange1;
			else if (!strcmp(kw4, "rb2")) cpp.cf = (ColorFunction *) GetRBRange2;
			else { PrintToLog("ERROR : add plane : unknown colour function.\n"); return; }
			
			f64 cscale = strtod(kw5, endptr);
			
			f64 dim = strtod(kw6, endptr);
			
			i32s res = strtol(kw7, endptr, 10);
			if (res < 2) res = 2;
			
			i32s tp = strtol(kw8, endptr, 10);
			if (tp < 0) tp = 0; if (tp > 1) tp = 1;
			
			f64 alpha = strtod(kw9, endptr);
			
			cpp.value = cscale; cpp.dim = dim; cpp.np = res;
			cpp.transparent = tp; cpp.alpha = alpha;
			
			ostrstream strN(object_name, sizeof(object_name));
			strN << kw3 << "-" << ends;
			
			AddObject(new color_plane_object(ol_static(), cpp, object_name));
			UpdateAllGraphicsViews();
			
			ostrstream strR(mbuff1, sizeof(mbuff1));
			strR << "Added a new object : plane (" << kw3 << " " << kw4 << ")." << endl << ends;
			PrintToLog(mbuff1);
			return;
		}
		
		if (!strcmp("volrend", kw2))
		{
			char kw3[32]; istr >> kw3;
			char kw4[32]; istr >> kw4;
			char kw5[32]; istr >> kw5;
			char kw6[32]; istr >> kw6;
			char kw7[32]; istr >> kw7;
			char kw8[32]; istr >> kw8;
			char ** endptr = NULL;
			
			cp_param cpp;
			cpp.prj = this; cpp.ref = GetCurrentSetup()->GetCurrentEngine();
			
			if (!strcmp(kw3, "esp")) cpp.vf = (ValueFunction *) value_ESP;
			else if (!strcmp(kw3, "vdws")) cpp.vf = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw3, "eldens")) cpp.vf = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw3, "mo")) cpp.vf = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw3, "mod")) cpp.vf = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw3, "unity")) cpp.vf = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add volrend : unknown value function.\n"); return; }
			
			if (!strcmp(kw4, "red")) cpp.cf = (ColorFunction *) GetRedColor;
			else if (!strcmp(kw4, "green")) cpp.cf = (ColorFunction *) GetGreenColor;
			else if (!strcmp(kw4, "blue")) cpp.cf = (ColorFunction *) GetBlueColor;
			else if (!strcmp(kw4, "rb1")) cpp.cf = (ColorFunction *) GetRBRange1;
			else if (!strcmp(kw4, "rb2")) cpp.cf = (ColorFunction *) GetRBRange2;
			else { PrintToLog("ERROR : add volrend : unknown colour function.\n"); return; }
			
			f64 cscale = strtod(kw5, endptr);
			
			f64 dim = strtod(kw6, endptr);
			
			i32s res = strtol(kw7, endptr, 10);
			if (res < 4) res = 4;
			
			f64 alpha = strtod(kw8, endptr);
			
			cpp.value = cscale; cpp.dim = dim; cpp.np = res;
			cpp.transparent = true; cpp.alpha = alpha;
			
			ostrstream strN(object_name, sizeof(object_name));
			strN << kw3 << "-" << ends;
			
			AddObject(new volume_rendering_object(ol_static(), cpp, res / 2, dim / 2.0, (* gv->cam), object_name));
			UpdateAllGraphicsViews();
			
			ostrstream strR(mbuff1, sizeof(mbuff1));
			strR << "Added a new object : volrend (" << kw3 << " " << kw4 << ")." << endl << ends;
			PrintToLog(mbuff1);
			return;
		}
		
		if (!strcmp("surf1", kw2))
		{
			char kw3[32]; istr >> kw3;	// vf1
			char kw4[32]; istr >> kw4;	// vf2
			char kw5[32]; istr >> kw5;	// cf
			char kw6[32]; istr >> kw6;	// sscale
			char kw7[32]; istr >> kw7;	// cscale
			char kw8[32]; istr >> kw8;	// dim
			char kw9[32]; istr >> kw9;	// res
			char kwA[32]; istr >> kwA;	// solid
			char kwB[32]; istr >> kwB;	// tp
			char kwC[32]; istr >> kwC;	// alpha
			char ** endptr = NULL;
			
			cs_param csp1;

			csp1.prj = this; csp1.ref = GetCurrentSetup()->GetCurrentEngine(); csp1.next = NULL;
			
			if (!strcmp(kw3, "esp")) csp1.vf1 = (ValueFunction *) value_ESP;
			else if (!strcmp(kw3, "vdws")) csp1.vf1 = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw3, "eldens")) csp1.vf1 = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw3, "mo")) csp1.vf1 = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw3, "mod")) csp1.vf1 = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw3, "unity")) csp1.vf1 = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add surf1 : unknown value function 1.\n"); return; }
			
			if (!strcmp(kw4, "esp")) csp1.vf2 = (ValueFunction *) value_ESP;
			else if (!strcmp(kw4, "vdws")) csp1.vf2 = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw4, "eldens")) csp1.vf2 = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw4, "mo")) csp1.vf2 = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw4, "mod")) csp1.vf2 = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw4, "unity")) csp1.vf2 = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add surf1 : unknown value function 2.\n"); return; }
			
			if (!strcmp(kw5, "red")) csp1.cf = (ColorFunction *) GetRedColor;
			else if (!strcmp(kw5, "green")) csp1.cf = (ColorFunction *) GetGreenColor;
			else if (!strcmp(kw5, "blue")) csp1.cf = (ColorFunction *) GetBlueColor;
			else if (!strcmp(kw5, "rb1")) csp1.cf = (ColorFunction *) GetRBRange1;
			else if (!strcmp(kw5, "rb2")) csp1.cf = (ColorFunction *) GetRBRange2;
			else { PrintToLog("ERROR : add surf1 : unknown colour function.\n"); return; }
			
			f64 sscale = strtod(kw6, endptr);
			f64 cscale = strtod(kw7, endptr);
			
			f64 dim = strtod(kw8, endptr);
			
			i32s res = strtol(kw9, endptr, 10);
			if (res < 4) res = 4;
			
			i32s solid = strtol(kwA, endptr, 10);
			if (solid < 0) solid = 0; if (solid > 1) solid = 1;

			i32s tp = strtol(kwB, endptr, 10);
			if (tp < 0) tp = 0; if (tp > 1) tp = 1;
			
			f64 alpha = strtod(kwC, endptr);
			
			static fGL dim_arr[3];
			dim_arr[0] = dim_arr[1] = dim_arr[2] = dim;
			
			static i32s res_arr[3];
			res_arr[0] = res_arr[1] = res_arr[2] = res;
			
			csp1.svalue = sscale; csp1.cvalue = cscale;
			csp1.dim = dim_arr; csp1.np = res_arr; csp1.wireframe = !solid;
			csp1.transparent = tp; csp1.alpha = alpha; csp1.toler = fabs(1.0e-6 * sscale); csp1.maxc = 250;
			
			ostrstream strN(object_name, sizeof(object_name));
			strN << kw3 << "-" << kw4 << "-" << ends;
			
			AddObject(new color_surface_object(ol_static(), csp1, object_name));
			UpdateAllGraphicsViews();
			
			ostrstream strR(mbuff1, sizeof(mbuff1));
			strR << "Added a new object : surf1 (" << kw3 << " " << kw4 << " " << kw5 << ")." << endl << ends;
			PrintToLog(mbuff1);
			return;
		}
		
		if (!strcmp("surf2", kw2))
		{
			char kw3[32]; istr >> kw3;	// vf1
			char kw4[32]; istr >> kw4;	// vf2
			char kw5[32]; istr >> kw5;	// cf1
			char kw6[32]; istr >> kw6;	// cf2
			char kw7[32]; istr >> kw7;	// sscale1
			char kw8[32]; istr >> kw8;	// sscale2
			char kw9[32]; istr >> kw9;	// cscale
			char kwA[32]; istr >> kwA;	// dim
			char kwB[32]; istr >> kwB;	// res
			char kwC[32]; istr >> kwC;	// solid
			char kwD[32]; istr >> kwD;	// tp
			char kwE[32]; istr >> kwE;	// alpha
			char ** endptr = NULL;
			
			cs_param csp2a; cs_param csp2b;
			
			csp2a.prj = this; csp2a.ref = GetCurrentSetup()->GetCurrentEngine(); csp2a.next = & csp2b;

			csp2b.prj = this; csp2b.ref = GetCurrentSetup()->GetCurrentEngine(); csp2b.next = NULL;
			
			if (!strcmp(kw3, "esp")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) value_ESP;
			else if (!strcmp(kw3, "vdws")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw3, "eldens")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw3, "mo")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw3, "mod")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw3, "unity")) csp2a.vf1 = csp2b.vf1 = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add surf2 : unknown value function 1.\n"); return; }
			
			if (!strcmp(kw4, "esp")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) value_ESP;
			else if (!strcmp(kw4, "vdws")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) value_VDWSurf;
			else if (!strcmp(kw4, "eldens")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) value_ElDens;
			else if (!strcmp(kw4, "mo")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) value_Orbital;
			else if (!strcmp(kw4, "mod")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) value_OrbDens;
			else if (!strcmp(kw4, "unity")) csp2a.vf2 = csp2b.vf2 = (ValueFunction *) GetUnity;
			else { PrintToLog("ERROR : add surf2 : unknown value function 2.\n"); return; }
			
			if (!strcmp(kw5, "red")) csp2a.cf = (ColorFunction *) GetRedColor;
			else if (!strcmp(kw5, "green")) csp2a.cf = (ColorFunction *) GetGreenColor;
			else if (!strcmp(kw5, "blue")) csp2a.cf = (ColorFunction *) GetBlueColor;
			else if (!strcmp(kw5, "rb1")) csp2a.cf = (ColorFunction *) GetRBRange1;
			else if (!strcmp(kw5, "rb2")) csp2a.cf = (ColorFunction *) GetRBRange2;
			else { PrintToLog("ERROR : add surf2 : unknown colour function 1.\n"); return; }
			
			if (!strcmp(kw6, "red")) csp2b.cf = (ColorFunction *) GetRedColor;
			else if (!strcmp(kw6, "green")) csp2b.cf = (ColorFunction *) GetGreenColor;
			else if (!strcmp(kw6, "blue")) csp2b.cf = (ColorFunction *) GetBlueColor;
			else if (!strcmp(kw6, "rb1")) csp2b.cf = (ColorFunction *) GetRBRange1;
			else if (!strcmp(kw6, "rb2")) csp2b.cf = (ColorFunction *) GetRBRange2;
			else { PrintToLog("ERROR : add surf2 : unknown colour function 2.\n"); return; }
			
			f64 sscale1 = strtod(kw7, endptr);
			f64 sscale2 = strtod(kw8, endptr);
			f64 cscale = strtod(kw9, endptr);
			
			f64 dim = strtod(kwA, endptr);
			
			i32s res = strtol(kwB, endptr, 10);
			if (res < 4) res = 4;
			
			i32s solid = strtol(kwC, endptr, 10);
			if (solid < 0) solid = 0; if (solid > 1) solid = 1;

			i32s tp = strtol(kwD, endptr, 10);
			if (tp < 0) tp = 0; if (tp > 1) tp = 1;
			
			f64 alpha = strtod(kwE, endptr);
			
			static fGL dim_arr[3];
			dim_arr[0] = dim_arr[1] = dim_arr[2] = dim;
			
			static i32s res_arr[3];
			res_arr[0] = res_arr[1] = res_arr[2] = res;
			
			csp2a.svalue = sscale1; csp2a.cvalue = cscale;
			csp2a.dim = dim_arr; csp2a.np = res_arr; csp2a.wireframe = !solid;
			csp2a.transparent = tp; csp2a.alpha = alpha; csp2a.toler = fabs(1.0e-6 * sscale1); csp2a.maxc = 250;
			
			csp2b.svalue = sscale2; csp2b.cvalue = cscale;
			csp2b.dim = dim_arr; csp2b.np = res_arr; csp2b.wireframe = !solid;
			csp2b.transparent = tp; csp2b.alpha = alpha; csp2b.toler = fabs(1.0e-6 * sscale2); csp2b.maxc = 250;
			
			ostrstream strN(object_name, sizeof(object_name));
			strN << kw3 << "-" << kw4 << "-" << ends;
			
			AddObject(new color_surface_object(ol_static(), csp2a, object_name));
			UpdateAllGraphicsViews();
			
			ostrstream strR(mbuff1, sizeof(mbuff1));
			strR << "Added a new object : surf2 (" << kw3 << " " << kw4 << " " << kw5 << " " << kw6 << ")." << endl << ends;
			PrintToLog(mbuff1);
			return;
		}
		
		PrintToLog("ERROR : could not process a command : \"add\".\n");
		return;
	}
	
	if (!strcmp("random_search", kw1))
	{
		char kw2[32]; istr >> kw2;	// the 2nd keyword; cycles.
		char kw3[32]; istr >> kw3;	// the 3rd keyword; optsteps.
		
		char ** endptr = NULL;
		i32s cycles = strtol(kw2, endptr, 10);
		i32s optsteps = strtol(kw3, endptr, 10);
		
		DoRandomSearch(cycles, optsteps);
		return;
	}

	if (!strcmp("systematic_search", kw1))
	{
		char kw2[32]; istr >> kw2;	// the 2nd keyword; divisions.
		char kw3[32]; istr >> kw3;	// the 3rd keyword; optsteps.
		
		char ** endptr = NULL;
		i32s divisions = strtol(kw2, endptr, 10);
		i32s optsteps = strtol(kw3, endptr, 10);
		
		DoSystematicSearch(divisions, optsteps);
		return;
	}

	if (!strcmp("montecarlo_search", kw1))
	{
		char kw2[32]; istr >> kw2;	// the 2nd keyword; n_init_steps.
		char kw3[32]; istr >> kw3;	// the 3rd keyword; n_simul_steps.
		char kw4[32]; istr >> kw4;	// the 4th keyword; optsteps.
		
		char ** endptr = NULL;
		i32s n_init_steps = strtol(kw2, endptr, 10);
		i32s n_simul_steps = strtol(kw3, endptr, 10);
		i32s optsteps = strtol(kw4, endptr, 10);
		
		DoMonteCarloSearch(n_init_steps, n_simul_steps, optsteps);
		return;
	}
	
	if (!strcmp("make_plot1", kw1))
	{
		char kw2[32]; istr >> kw2;	// A
		char kw3[32]; istr >> kw3;	// B
		char kw4[32]; istr >> kw4;	// C
		char kw5[32]; istr >> kw5;	// D
		char kw6[32]; istr >> kw6;	// div
		char kw7[32]; istr >> kw7;	// start_ang
		char kw8[32]; istr >> kw8;	// end_ang
		char kw9[32]; istr >> kw9;	// optsteps
		char ** endptr = NULL;
		
		i32s ia = strtol(kw2, endptr, 10);
		i32s ib = strtol(kw3, endptr, 10);
		i32s ic = strtol(kw4, endptr, 10);
		i32s id = strtol(kw5, endptr, 10);
		i32s div1 = strtol(kw6, endptr, 10);
		fGL start1 = strtod(kw7, endptr);
		fGL end1 = strtod(kw8, endptr);
		
		i32s optsteps = strtol(kw9, endptr, 10);
		
		DoEnergyPlot1D(ia, ib, ic, id, div1, start1, end1, optsteps);
		return;
	}
	
	if (!strcmp("make_plot2", kw1))
	{
		char kw2[32]; istr >> kw2;	// A
		char kw3[32]; istr >> kw3;	// B
		char kw4[32]; istr >> kw4;	// C
		char kw5[32]; istr >> kw5;	// D
		char kw6[32]; istr >> kw6;	// div
		char kw7[32]; istr >> kw7;	// start_ang
		char kw8[32]; istr >> kw8;	// end_ang
		char kw9[32]; istr >> kw9;	// I
		char kwA[32]; istr >> kwA;	// J
		char kwB[32]; istr >> kwB;	// K
		char kwC[32]; istr >> kwC;	// L
		char kwD[32]; istr >> kwD;	// div
		char kwE[32]; istr >> kwE;	// start_ang
		char kwF[32]; istr >> kwF;	// end_ang
		char kwG[32]; istr >> kwG;	// optsteps
		char ** endptr = NULL;
		
		i32s ia = strtol(kw2, endptr, 10);
		i32s ib = strtol(kw3, endptr, 10);
		i32s ic = strtol(kw4, endptr, 10);
		i32s id = strtol(kw5, endptr, 10);
		i32s div1 = strtol(kw6, endptr, 10);
		fGL start1 = strtod(kw7, endptr);
		fGL end1 = strtod(kw8, endptr);
		
		i32s ii = strtol(kw9, endptr, 10);
		i32s ij = strtol(kwA, endptr, 10);
		i32s ik = strtol(kwB, endptr, 10);
		i32s il = strtol(kwC, endptr, 10);
		i32s div2 = strtol(kwD, endptr, 10);
		fGL start2 = strtod(kwE, endptr);
		fGL end2 = strtod(kwF, endptr);
		
		i32s optsteps = strtol(kwG, endptr, 10);
		
		DoEnergyPlot2D(ia, ib, ic, id, div1, start1, end1, ii, ij, ik, il, div2, start2, end2, optsteps);
		return;
	}
	
	if (!strcmp("set_current_orbital", kw1))
	{
		char kw2[32]; istr >> kw2;	// the 2nd keyword; the orbital index.
		
		char ** endptr = NULL;
		int index = strtol(kw2, endptr, 10);
		if (index < 0) index = 0;
		
		qm_current_orbital = index;
		
		ostrstream strR(mbuff1, sizeof(mbuff1));
		strR << "The current orbital is now " << qm_current_orbital << "." << endl << ends;
		PrintToLog(mbuff1);
		return;
	}
	
	if (!strcmp("build_amino", kw1))
	{
		char kw2[4096]; istr >> kw2;	// sequence
		char kw3[32]; istr >> kw3;	// helix/sheet (optional)
		
		f64 aab_ah[] = { 302.0 * M_PI / 180.0, 313.0 * M_PI / 180.0, M_PI };
		f64 aab_bs[] = { 180.0 * M_PI / 180.0, 180.0 * M_PI / 180.0, M_PI };
		
		model::amino_builder.Build(this, kw2, kw3[0] == 'h' || kw3[0] == 'H' ? aab_ah : aab_bs);
		UpdateAllGraphicsViews();
		
		ostrstream strR(mbuff1, sizeof(mbuff1));
		strR << "built a sequence : " << kw2 << endl << ends;
		PrintToLog(mbuff1);
		return;
	}
	
	if (!strcmp("build_nucleic", kw1))
	{
		char kw2[4096]; istr >> kw2;	// sequence
		
		f64 nab[] =
		{
			44.6 * M_PI / 180.0, 261.0 * M_PI / 180.0, 320.8 * M_PI / 180.0, 208.6 * M_PI / 180.0,
			273.8 * M_PI / 180.0, 105.6 * M_PI / 180.0, 356.0 * M_PI / 180.0, 24.7 * M_PI / 180.0,
			88.7 * M_PI / 180.0, 264.6 * M_PI / 180.0
		};
		
		model::nucleic_builder.Build(this, kw2, nab);
		UpdateAllGraphicsViews();
		
		ostrstream strR(mbuff1, sizeof(mbuff1));
		strR << "built a sequence : " << kw2 << endl << ends;
		PrintToLog(mbuff1);
		return;
	}
	
	// if the command is not recognized above, we will print out an error message here.
	
	ostrstream strE(mbuff1, sizeof(mbuff1)); strE << "ERROR : Unknown command : " << command << endl << ends;
	PrintToLog(mbuff1); PrintToLog("The \"help\" command will give more information about command strings.\n");
}

void project::DoDeleteCurrentObject(void)
{
	if (selected_object != NULL)
	{
		bool test1 = RemoveLight(selected_object);
		bool test2 = test1; if (!test1) test2 = RemoveObject(selected_object);
		
		if (test2)
		{
			selected_object = NULL;
			event_SelectedObjectChanged();
			UpdateAllGraphicsViews();
		}
	}
}

void project::DoSwitchLocalLights(camera * cam, bool report)
{
	cam->use_local_lights = !cam->use_local_lights;
	if (report) cout << "local lights = " << (cam->use_local_lights ? "on" : "off") << endl;
	SetupLights(cam); UpdateGraphicsViews(cam);
}

void project::DoSwitchGlobalLights(camera * cam, bool report)
{
	cam->use_global_lights = !cam->use_global_lights;
	if (report) cout << "global lights = " << (cam->use_global_lights ? "on" : "off") << endl;
	SetupLights(cam); UpdateGraphicsViews(cam);
}


void project::InitGL(void)
{
        const fGL default_background[4] = { 0.0, 0.0, 0.0, 1.0};
	fGL *background = model_prefs->ColorRGBA("Graphics/BkgndColor", default_background);
	glClearColor(background[0], background[1], background[2], background[3]);
	delete [] background;
	
	glDepthFunc(GL_LESS); glEnable(GL_DEPTH_TEST);
	glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 64);
	
	const fGL default_reflectance[4] = { 0.9, 0.9, 0.9, 1.0 };
	fGL *specular_reflectance = model_prefs->ColorRGBA("Graphics/SpecularReflectance", default_reflectance);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular_reflectance);
	delete [] specular_reflectance;
	
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	
	const fGL default_ambient[4] = { 0.2, 0.2, 0.2, 1.0 };
	fGL *ambient_intensity = model_prefs->ColorRGBA("Graphics/AmbientIntensity", default_ambient);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_intensity);
	delete [] ambient_intensity;
	
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, true);
	
	glFogi(GL_FOG_MODE, GL_EXP);
	glFogf(GL_FOG_DENSITY, model_prefs->Double("Graphics/FogDensity",0.15));
	
	const fGL default_fog[4] = { 0.0, 0.0, 0.0, 0.0 };
	fGL *fog_color = model_prefs->ColorRGBA("Graphics/FogColor", default_fog);
	glFogfv(GL_FOG_COLOR, fog_color);
	delete [] fog_color;
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glSelectBuffer(SB_SIZE, select_buffer);
}

fGL project::GetDefaultFocus(void)
{
	return 2.0;
}

const char * project::GetType(void)
{
	return "MM, all-atoms model";
}

color_mode * project::GetDefaultColorMode(void)
{
	return & project::cm_element;
}

void project::SelectAll(void)
{
	if (selected_object != NULL)
	{
		selected_object = NULL;
		event_SelectedObjectChanged();
	}
	
	iter_al it1 = atom_list.begin();
	while (it1 != atom_list.end()) (* it1++).flags |= ATOMFLAG_SELECTED;
	
	UpdateAllGraphicsViews();
}

void project::InvertSelection(void)
{
	if (selected_object != NULL)
	{
		selected_object = NULL;
		event_SelectedObjectChanged();
	}
	
	iter_al it1 = atom_list.begin();
	while (it1 != atom_list.end()) (* it1++).flags ^= ATOMFLAG_SELECTED;
	
	UpdateAllGraphicsViews();
}

bool project::TestAtom(atom * ref, rmode rm)
{
	// what about the "visible"-flag???
	
	if (rm == Transform1 && (ref->flags & ATOMFLAG_SELECTED)) return false;
	if (rm == Transform2 && !(ref->flags & ATOMFLAG_SELECTED)) return false;
	
	return true;
}

bool project::TestBond(bond * ref, rmode rm)
{
	// what about the "visible"-flag???
	
	if (rm == Transform1 && (ref->atmr[0]->flags & ATOMFLAG_SELECTED)) return false;
	if (rm == Transform2 && !(ref->atmr[0]->flags & ATOMFLAG_SELECTED)) return false;
	
	bool test1 = (ref->atmr[0]->flags & ATOMFLAG_SELECTED) ? true : false;
	bool test2 = (ref->atmr[1]->flags & ATOMFLAG_SELECTED) ? true : false;
	if (rm != Normal && test1 != test2) return false;
	
	return true;
}

void project::SetColor(color_mode * cm, atom * ref)
{
	const fGL default_sel_color[3] = { 1.0, 0.0, 1.0 };
	fGL *select_color = model_prefs->ColorRGB("Graphics/SelectColor", default_sel_color);

	if (ref->flags & ATOMFLAG_SELECTED)
	{
		glColor3f(select_color[0], select_color[1], select_color[2]);
	}
	else
	{
		static fGL color[4];
		cm->GetColor4(ref, -1, color);
		glColor3fv(color);
	}

	delete [] select_color;
}

void project::Render(graphics_view * gv, rmode rm)
{
	const fGL default_label_color[3] = { 0.0, 1.0, 1.0 };	// looks bad but won't fade easily into other colours...
	fGL *label_color = model_prefs->ColorRGB("Graphics/LabelColor", default_label_color);
	
	bool accum = gv->accumulate; if (rm != Normal) accum = false;
//if (accum) { glClear(GL_ACCUM_BUFFER_BIT); UpdateAccumValues(); }
//else if (rm != Transform2) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (periodic && rm == Normal)
	{
		glLineWidth(1.0);
		glColor3f(1.0, 0.0, 1.0);
		glBegin(GL_LINES);
		
		glVertex3f(-box_hdim[0], -box_hdim[1], +box_hdim[2]);
		glVertex3f(+box_hdim[0], -box_hdim[1], +box_hdim[2]);
		
		glVertex3f(-box_hdim[0], +box_hdim[1], +box_hdim[2]);
		glVertex3f(+box_hdim[0], +box_hdim[1], +box_hdim[2]);
		
		glVertex3f(-box_hdim[0], -box_hdim[1], -box_hdim[2]);
		glVertex3f(+box_hdim[0], -box_hdim[1], -box_hdim[2]);
		
		glVertex3f(-box_hdim[0], +box_hdim[1], -box_hdim[2]);
		glVertex3f(+box_hdim[0], +box_hdim[1], -box_hdim[2]);
		
		glVertex3f(-box_hdim[0], -box_hdim[1], -box_hdim[2]);
		glVertex3f(-box_hdim[0], -box_hdim[1], +box_hdim[2]);
		
		glVertex3f(-box_hdim[0], +box_hdim[1], -box_hdim[2]);
		glVertex3f(-box_hdim[0], +box_hdim[1], +box_hdim[2]);
		
		glVertex3f(+box_hdim[0], -box_hdim[1], -box_hdim[2]);
		glVertex3f(+box_hdim[0], -box_hdim[1], +box_hdim[2]);
		
		glVertex3f(+box_hdim[0], +box_hdim[1], -box_hdim[2]);
		glVertex3f(+box_hdim[0], +box_hdim[1], +box_hdim[2]);
		
		glVertex3f(-box_hdim[0], -box_hdim[1], -box_hdim[2]);
		glVertex3f(-box_hdim[0], +box_hdim[1], -box_hdim[2]);
		
		glVertex3f(-box_hdim[0], -box_hdim[1], +box_hdim[2]);
		glVertex3f(-box_hdim[0], +box_hdim[1], +box_hdim[2]);
		
		glVertex3f(+box_hdim[0], -box_hdim[1], -box_hdim[2]);
		glVertex3f(+box_hdim[0], +box_hdim[1], -box_hdim[2]);
		
		glVertex3f(+box_hdim[0], -box_hdim[1], +box_hdim[2]);
		glVertex3f(+box_hdim[0], +box_hdim[1], +box_hdim[2]);
		
		glEnd();
	}
	
	if (gv->enable_fog) glEnable(GL_FOG);
	
	i32s layers = 0;
	if (periodic && rm == Normal) layers = 1;
	
	for (i32s r1 = -layers;r1 < (layers + 1);r1++)
	{
		for (i32s r2 = -layers;r2 < (layers + 1);r2++)
		{
			for (i32s r3 = -layers;r3 < (layers + 1);r3++)
			{
				glPushMatrix();
				
				fGL trans1 = r1 * box_fdim[0];
				fGL trans2 = r2 * box_fdim[1];
				fGL trans3 = r3 * box_fdim[2];
				
				glTranslated(trans1, trans2, trans3);
				
				RenderScene(gv, rm, accum);
				
				glPopMatrix();
			}
		}
	}
	
	if (accum) glAccum(GL_RETURN, 1.0);
	else if (rm != Transform2) gv->cam->RenderObjects(gv);
	
	if (gv->label == LABEL_INDEX)
	{
		char string[32]; i32s tmp1 = 0;
		
		glColor3f(label_color[0], label_color[1], label_color[2]);
		for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
		{
			ostrstream str(string, sizeof(string));
			str << tmp1++ << ends;
			
			const fGL * cdata = (* it1).GetCRD(0);
			fGL x = cdata[0]; fGL y = cdata[1]; fGL z = cdata[2];
			
			gv->WriteGlutString3D(string, x, y, z, gv->cam->GetLocData(), GLUT_BITMAP_9_BY_15);
		}
	}
	else if (gv->label == LABEL_CHARGE)
	{
		char string[32];
		
		glColor3f(label_color[0], label_color[1], label_color[2]);
		for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
		{
			ostrstream str(string, sizeof(string));
			str.setf(ios::fixed | ios::showpos); str.precision(4); str << (* it1).charge << ends;

			const fGL * cdata = (* it1).GetCRD(0);
			fGL x = cdata[0]; fGL y = cdata[1]; fGL z = cdata[2];
			
			gv->WriteGlutString3D(string, x, y, z, gv->cam->GetLocData(), GLUT_BITMAP_9_BY_15);
		}
	}
	else if (gv->label == LABEL_ELEMENT)
	{
		char string[32];
		
		glColor3f(label_color[0], label_color[1], label_color[2]);
		for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
		{
			ostrstream str(string, sizeof(string));
			str << (* it1).el.GetSymbol() << ends;

			const fGL * cdata = (* it1).GetCRD(0);
			fGL x = cdata[0]; fGL y = cdata[1]; fGL z = cdata[2];
			
			gv->WriteGlutString3D(string, x, y, z, gv->cam->GetLocData(), GLUT_BITMAP_9_BY_15);
		}
	}
	else if (gv->label == LABEL_ATOMTYPE)
	{
		char string[32];
		
		glColor3f(label_color[0], label_color[1], label_color[2]);
		for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
		{
			ostrstream str(string, sizeof(string));
			str << "0x" << hex << (* it1).atmtp << ends;

			const fGL * cdata = (* it1).GetCRD(0);
			fGL x = cdata[0]; fGL y = cdata[1]; fGL z = cdata[2];
			
			gv->WriteGlutString3D(string, x, y, z, gv->cam->GetLocData(), GLUT_BITMAP_9_BY_15);
		}
	}
	else if (gv->label == LABEL_BONDTYPE)
	{
		char string[32];
		
		glColor3f(label_color[0], label_color[1], label_color[2]);
		for (iter_bl it1 = bond_list.begin();it1 != bond_list.end();it1++)
		{
			ostrstream str(string, sizeof(string));
			str << (* it1).bt.GetSymbol1() << ends;
			
			const fGL * cd1 = (* it1).atmr[0]->GetCRD(0); const fGL * cd2 = (* it1).atmr[1]->GetCRD(0);
			fGL x = (cd1[0] + cd2[0]) / 2.0; fGL y = (cd1[1] + cd2[1]) / 2.0; fGL z = (cd1[2] + cd2[2]) / 2.0;
			
			gv->WriteGlutString3D(string, x, y, z, gv->cam->GetLocData(), GLUT_BITMAP_9_BY_15);
		}
	}
	
	if (gv->enable_fog) glDisable(GL_FOG);
	
	// finally call this to handle transparency...
	// finally call this to handle transparency...
	// finally call this to handle transparency...
	
	RenderAllTPs(gv, rm);

	delete [] label_color;
}

void project::RenderScene(graphics_view * gv, rmode rm, bool accum)
{
	for (i32u n1 = 0;n1 < cs_vector.size();n1++)
	{
		if (!GetCRDSetVisible(n1)) continue;
if (accum) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// FIXME!!!

		if (gv->render == RENDER_WIREFRAME)
		{
			glPointSize(3.0); glLineWidth(1.0);
			for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)		// wireframe atoms
			{
				if (!TestAtom(& (* it1), rm)) continue;
				glPushName(GLNAME_MD_TYPE1); glPushName((i32u) & (* it1));
				
				glBegin(GL_POINTS);
				SetColor(gv->colormode, & (* it1));
				glVertex3fv((* it1).GetCRD(n1));
				glEnd();
				
				glPopName(); glPopName();
			}
			
			glEnable(GL_LINE_STIPPLE);
			for (iter_bl it2 = bond_list.begin();it2 != bond_list.end();it2++)		// wireframe bonds
			{
				if (!TestBond(& (* it2), rm)) continue;
				
				switch ((* it2).bt.GetSymbol1())
				{
					case 'S': glLineStipple(1, 0xFFFF); break;
					case 'C': glLineStipple(1, 0x3FFF); break;
					case 'D': glLineStipple(1, 0x3F3F); break;
					case 'T': glLineStipple(1, 0x3333); break;
				}
				
				glBegin(GL_LINES);
				SetColor(gv->colormode, (* it2).atmr[0]);
				glVertex3fv((* it2).atmr[0]->GetCRD(n1));
				SetColor(gv->colormode, (* it2).atmr[1]);
				glVertex3fv((* it2).atmr[1]->GetCRD(n1));
				glEnd();
			}
			glDisable(GL_LINE_STIPPLE);
		}
		
		if (gv->render != RENDER_WIREFRAME && gv->render != RENDER_NOTHING)
		{
			glEnable(GL_LIGHTING);
			
			for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)		// atoms as spheres
			{
				if (!TestAtom(& (* it1), rm)) continue;
				
				SetColor(gv->colormode, & (* it1));
				
				float rad = 0.0; int res = 0;
				switch (gv->render)
				{
					case RENDER_BALL_AND_STICK:
					rad = model_prefs->Double("MM1Graphics/BallSize", 0.035);
					if (model_prefs->Boolean("MM1Graphics/BallVdWScale", false))
					    rad *= (* it1).vdwr * 4.0;
					res = model_prefs->Value("MM1Graphics/BallResolution", 12);
					break;
					
					case RENDER_VAN_DER_WAALS:
					rad = (* it1).vdwr;
					res = model_prefs->Value("MM1Graphics/VdWResolution", 22);
					break;
					
					case RENDER_CYLINDERS:
					rad = model_prefs->Double("MM1Graphics/CylinderSize", 0.035);
					res = model_prefs->Value("MM1Graphics/CylinderResolution", 12);
					break;
				}
				
				glPushName(GLNAME_MD_TYPE1); glPushName((i32u) & (* it1));
				
				GLUquadricObj * qo = gluNewQuadric();
				gluQuadricDrawStyle(qo, (GLenum) GLU_FILL);
				
				glPushMatrix();
				const fGL * cdata = (* it1).GetCRD(n1);
				glTranslated(cdata[0], cdata[1], cdata[2]);
				gluSphere(qo, rad, res, res / 2);
				glPopMatrix();
				gluDeleteQuadric(qo);
				
				glPopName(); glPopName();
			}
			
			glDisable(GL_LIGHTING);
		}
		
		if (gv->render == RENDER_BALL_AND_STICK || gv->render == RENDER_CYLINDERS)
		{
			glEnable(GL_LIGHTING);
			
			for (iter_bl it1 = bond_list.begin();it1 != bond_list.end();it1++)		// bonds as cylinders
			{
				if (!TestBond(& (* it1), rm)) continue;
				
				fGL vdwr[2] =
				{
					(* it1).atmr[0]->vdwr,
					(* it1).atmr[1]->vdwr
				};
				
				fGL vdwrsum = vdwr[0] + vdwr[1];
				
				for (i32s n2 = 0;n2 < 2;n2++)
				{
					const fGL * crd1 = (* it1).atmr[n2]->GetCRD(n1);
					const fGL * crd2 = (* it1).atmr[!n2]->GetCRD(n1);
					v3d<fGL> crt1 = v3d<fGL>(crd1);
					v3d<fGL> crt2 = v3d<fGL>(crd2);
					v3d<fGL> crt = crt2 - crt1;
					
					fGL pol[3]; crt2pol(crt.data, pol);
					
					SetColor(gv->colormode, (* it1).atmr[n2]);
					
					float trans, rad = 0.0; int res = 0;
					switch (gv->render)
					{
						case RENDER_BALL_AND_STICK:
						rad = model_prefs->Double("MM1Graphics/StickSize", 0.01);
						res = model_prefs->Value("MM1Graphics/StickResolution", 6);
						break;
						
						case RENDER_CYLINDERS:
						rad = model_prefs->Double("MM1Graphics/CylinderSize", 0.035);
						res = model_prefs->Value("MM1Graphics/CylinderResolution", 12);
						break;
					}
					
					glPushName(GLNAME_MD_TYPE1); glPushName((i32u) (* it1).atmr[n2]);
					
					GLUquadricObj * qo = gluNewQuadric();
					gluQuadricDrawStyle(qo, (GLenum) GLU_FILL);
					glPushMatrix();
					
					glTranslated(crd1[0], crd1[1], crd1[2]);
					
					glRotated(180.0 * pol[1] / M_PI, 0.0, 1.0, 0.0);
					glRotated(180.0 * pol[2] / M_PI, sin(-pol[1]), 0.0, cos(-pol[1]));
					
					// any chance to further define the orientation of, for example, double bonds???
					// one more rotation would be needed. but what is the axis, and how much to rotate???
					
					fGL length = crt.len() * vdwr[n2] / vdwrsum;
					
					if (gv->render == RENDER_BALL_AND_STICK)
					switch ((* it1).bt.GetValue())
					{
						case BONDTYPE_DOUBLE:
						trans = rad;
						rad = rad / 1.5;
						
						if (n2)
							glTranslated(0.0, trans, 0.0);
						else
							glTranslated(0.0, -trans, 0.0);
						gluCylinder(qo, rad, rad, length, res, 1);					
						if (n2)
							glTranslated(0.0, -2.0 * trans, 0.0);
						else
							glTranslated(0.0, 2.0 * trans, 0.0);
						gluCylinder(qo, rad, rad, length, res, 1);
						break;
						
						case BONDTYPE_CNJGTD:
						trans = rad;
						rad = rad / 1.5;
						
						if (n2)
							glTranslated(0.0, trans, 0.0);
						else
							glTranslated(0.0, -trans, 0.0);
						gluCylinder(qo, rad, rad, length, res, 1);
						if (n2)
							glTranslated(0.0, -2.0 * trans, 0.0);
						else
							glTranslated(0.0, 2.0 * trans, 0.0);
						
						glEnable(GL_LINE_STIPPLE);
						glLineStipple(1, 0x3F3F);
						gluQuadricDrawStyle(qo, (GLenum) GLU_LINE);
						gluCylinder(qo, rad, rad, length, res, 1);
						glDisable(GL_LINE_STIPPLE);
						break;
						
						case BONDTYPE_TRIPLE:
						trans = rad;
						rad = rad / 2.0;
						
						if (n2)
							glTranslated(0.0, trans, 0.0);
						else
							glTranslated(0.0, -trans, 0.0);
						gluCylinder(qo, rad, rad, length, res, 1);
						if (n2)
							glTranslated(0.0, -trans, 0.0);
						else
							glTranslated(0.0, trans, 0.0);
						gluCylinder(qo, rad, rad, length, res, 1);
						if (n2)
							glTranslated(0.0, -trans, 0.0);
						else
							glTranslated(0.0, trans, 0.0);
						gluCylinder(qo, rad, rad, length, res,1);
						break;
						
						default:
						gluCylinder(qo, rad, rad, length, res, 1);
					}
					else
						gluCylinder(qo, rad, rad, length, res, 1);
					
					glPopMatrix();
					gluDeleteQuadric(qo);
					
					glPopName(); glPopName();
				}
			}
			
			glDisable(GL_LIGHTING);
		}
		
		glEnable(GL_LIGHTING); glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, true); glBegin(GL_QUADS);
		for (iter_bl it1 = bond_list.begin();it1 != bond_list.end();it1++)	// sf peptide dipoles...
		{
			if ((* it1).sf_pbdd < -1000.0) continue;
			
		//char symbol2 = chn_vector[n2].res_vector[n3 + 1].symbol;
		//if (symbol2 == 'P') continue;	// skip all X-pro cases !!!
			
			atom * prev = NULL;
			atom * curr = (* it1).atmr[0];
			atom * next = (* it1).atmr[1];
			
			// WARNING!!! this is pretty slow!!! need to find the previous c-alpha.
			iter_cl it2;
			for (it2 = curr->cr_list.begin();it2 != curr->cr_list.end();it2++)
			{
				if ((* it2).atmr == next) continue;
				
				if ((* it2).atmr->el.GetAtomicNumber() > 0) continue;
				if ((* it2).atmr->sf_atmtp & 0xFF) continue;
				
				prev = (* it2).atmr;
				break;
			}
			
			if (!prev) continue;
			
			v3d<fGL> v1(prev->GetCRD(n1), curr->GetCRD(n1));
			v3d<fGL> v2(curr->GetCRD(n1), next->GetCRD(n1));
			
			v3d<fGL> v3 = v1.vpr(v2); v3 = v3 * (0.075 / v3.len());
			v3d<fGL> v4 = v3.vpr(v2); v4 = v4 * (0.075 / v4.len());
			
			fGL peptide = (* it1).sf_pbdd;	// this is the same for all crd_sets!!!
			v3d<fGL> v5 = (v3 * sin(peptide)) + (v4 * cos(peptide));
			
			fGL peptnorm = peptide - M_PI / 2.0;
			v3d<fGL> normal = (v3 * sin(peptnorm)) + (v4 * cos(peptnorm));
			normal = normal / normal.len(); glNormal3fv(normal.data);
			
			v3d<fGL> pvc(curr->GetCRD(n1));
			v3d<fGL> pv1 = pvc + (v2 * 0.5) + v5; v3d<fGL> pv2 = pvc + (v2 * 0.90);
			v3d<fGL> pv3 = pvc + (v2 * 0.5) - v5; v3d<fGL> pv4 = pvc + (v2 * 0.10);
			
			glColor3f(1.0, 0.0, 0.0); glVertex3fv(pv1.data);
			glColor3f(0.0, 1.0, 0.0); glVertex3fv(pv2.data);
			glColor3f(0.0, 0.0, 1.0); glVertex3fv(pv3.data);
			glColor3f(0.0, 1.0, 0.0); glVertex3fv(pv4.data);
		}
		glEnd();	// GL_QUADS
		glDisable(GL_LIGHTING); glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, false);
		
		if (accum)
		{
			gv->cam->RenderObjects(gv);
			glAccum(GL_ACCUM, cs_vector[n1]->accum_value);
		}
	}
}

void project::Center(transformer * p1)
{
	i32s sum = 0;
	p1->GetLocDataRW()->crd[0] = 0.0;
	p1->GetLocDataRW()->crd[1] = 0.0;
	p1->GetLocDataRW()->crd[2] = 0.0;
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		if (!((* it1).flags & ATOMFLAG_SELECTED)) continue;
		for (i32u n1 = 0;n1 < cs_vector.size();n1++)
		{
			sum++;
			const fGL * cdata = (* it1).GetCRD(n1);
			p1->GetLocDataRW()->crd[0] += cdata[0];
			p1->GetLocDataRW()->crd[1] += cdata[1];
			p1->GetLocDataRW()->crd[2] += cdata[2];
		}
	}
	
	if (!sum) return;
	
	p1->GetLocDataRW()->crd[0] /= (fGL) sum;
	p1->GetLocDataRW()->crd[1] /= (fGL) sum;
	p1->GetLocDataRW()->crd[2] /= (fGL) sum;
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		if (!((* it1).flags & ATOMFLAG_SELECTED)) continue;
		for (i32u n1 = 0;n1 < cs_vector.size();n1++)
		{
			const fGL * cdata = (* it1).GetCRD(n1);
			
			fGL x = cdata[0] - p1->GetLocData()->crd[0];
			fGL y = cdata[1] - p1->GetLocData()->crd[1];
			fGL z = cdata[2] - p1->GetLocData()->crd[2];
			
			(* it1).SetCRD(n1, x, y, z);
		}
	}
}

void project::Transform(transformer * p1)
{
	fGL matrix[16]; p1->GetMatrix(matrix);
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		if (!((* it1).flags & ATOMFLAG_SELECTED)) continue;
		
		for (i32u n1 = 0;n1 < cs_vector.size();n1++)
		{
			v3d<fGL> posv = v3d<fGL>((* it1).GetCRD(n1));
			TransformVector(posv, matrix);
			
			(* it1).SetCRD(n1, posv[0], posv[1], posv[2]);
		}
	}
}

void project::DrawEvent(graphics_view * gv, vector<iGLu> & names)
{
	if (ogl_view::button == mouse_tool::Right) return;	// the right button is for popup menus...
	
	i32s mouse[2] =
	{
		gv->current_tool->latest_x,
		gv->current_tool->latest_y
	};
	
	if (ogl_view::state == mouse_tool::Down)
	{
		if (names.size() > 1 && names[0] == GLNAME_MD_TYPE1)
		{
			draw_data[0] = (atom *) names[1];
		}
		else
		{
			fGL tmp1[3]; gv->GetCRD(mouse, tmp1);
			atom newatom(element::current_element, tmp1, cs_vector.size());
			AddAtom(newatom); draw_data[0] = & atom_list.back();
		}
	}
	else
	{
		if (names.size() > 1 && names[0] == GLNAME_MD_TYPE1)
		{
			draw_data[1] = (atom *) names[1];
		}
		else
		{
			fGL tmp1[3]; gv->GetCRD(mouse, tmp1);
			atom newatom(element::current_element, tmp1, cs_vector.size());
			AddAtom(newatom); draw_data[1] = & atom_list.back();
		}
		
		// if different: update bondtype or add a new bond.
		// if not different: change atom to different element.
		
		if (draw_data[0] != draw_data[1])
		{
			bond newbond(draw_data[0], draw_data[1], bondtype::current_bondtype);
			iter_bl it1 = find(bond_list.begin(), bond_list.end(), newbond);
			if (it1 != bond_list.end()) (* it1).bt = bondtype::current_bondtype;
			else AddBond(newbond);
		}
		else
		{
			draw_data[0]->el = element::current_element;
		}
		
		UpdateAllGraphicsViews();
	}
}

void project::EraseEvent(graphics_view * gv, vector<iGLu> & names)
{
	if (ogl_view::button == mouse_tool::Right) return;	// the right button is for popup menus...
	
	if (ogl_view::state == mouse_tool::Down)
	{
		if (names.size() > 1 && names[0] == GLNAME_MD_TYPE1)
		{
			draw_data[0] = (atom *) names[1];
		}
		else
		{
			draw_data[0] = NULL;
		}
	}
	else
	{
		if (names.size() > 1 && names[0] == GLNAME_MD_TYPE1)
		{
			draw_data[1] = (atom *) names[1];
		}
		else
		{
			draw_data[1] = NULL;
		}
		
		if (!draw_data[0] || !draw_data[1]) return;
		
		// if different: try to find and remove a bond.
		// if not different: remove atom.
		
		if (draw_data[0] != draw_data[1])
		{
			bond tmpbond(draw_data[0], draw_data[1], bondtype::current_bondtype);
			iter_bl it1 = find(bond_list.begin(), bond_list.end(), tmpbond);
			if (it1 != bond_list.end()) RemoveBond(it1); else return;
		}
		else
		{
			iter_al it1 = find(atom_list.begin(), atom_list.end(), (* draw_data[0]));
			if (it1 != atom_list.end()) RemoveAtom(it1);
			else
			{
				cout << "BUG: project::EraseEvent()." << endl;
				exit(EXIT_FAILURE);
			}
		}
		
		UpdateAllGraphicsViews();
	}
}

void project::SelectEvent(graphics_view *, vector<iGLu> & names)
{
	if (names[0] == GLNAME_MD_TYPE1)
	{
		atom * ref = (atom *) names[1];
		ref->flags ^= ATOMFLAG_SELECTED;
		UpdateAllGraphicsViews();
		
		cout << "atomtype is " << hex << ref->atmtp << dec << endl;
	}
}

void project::MeasureEvent(graphics_view *, vector<iGLu> & names)
{
	char mbuff1[256];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	
        static atom *a1 = NULL;
	static atom *a2 = NULL;
	static atom *a3 = NULL;
// could this be shorter as a loop???

	if (names[0] == GLNAME_MD_TYPE1)
	{
		atom * ref = (atom *) names[1];
		ref->flags ^= ATOMFLAG_SELECTED;
		UpdateAllGraphicsViews();
		
		if (a1 == NULL)
		  {
		    a1 = ref;
		    str1 << "charge: " << ref->charge << endl << ends;
		    PrintToLog(mbuff1);
		  }
		else if (a1 != NULL && a2 == NULL)
		  {
		    if (a1 == ref) { a1->flags &= (~ATOMFLAG_SELECTED); a1 = NULL; return; }
		    
		    a2 = ref;
		    
		    const fGL * p1 = a1->GetCRD(0);
		    const fGL * p2 = a2->GetCRD(0);
		    
		    //float len = measure_len(p1, p2);			// not const-correct???
		    v3d<fGL> v1(p1, p2);
		    fGL len = v1.len();
		    
		    str1 << "distance: " << len << " nm" << endl << ends;
		    PrintToLog(mbuff1);
		  }
		else if (a1 != NULL && a2 != NULL && a3 == NULL)
		  {
		    if (a1 == ref) { a1->flags &= (~ATOMFLAG_SELECTED); a1 = a2; a2 = NULL; return; }
		    else if (a2 == ref) { a2->flags &= (~ATOMFLAG_SELECTED); a2 = NULL; return; }
		    
		    a3 = ref;
		    
		    const fGL * p1 = a1->GetCRD(0);
		    const fGL * p2 = a2->GetCRD(0);
		    const fGL * p3 = a3->GetCRD(0);
		    
		    //float ang = measure_ang(p1, p2, p3);		// not const-correct???
		    v3d<fGL> v1(p2, p1);
		    v3d<fGL> v2(p2, p3);
		    fGL ang = v1.ang(v2) * 180.0 / M_PI;
		    
		    str1 << "angle: " << ang << " deg" << endl << ends;
		    PrintToLog(mbuff1);
		  }
		else
		  {
		    if (a1 == ref) { a1->flags &= (~ATOMFLAG_SELECTED); a1 = a2; a2 = a3; a3 = NULL; return; }
		    else if (a2 == ref) { a2->flags &= (~ATOMFLAG_SELECTED); a2 = a3; a3 = NULL; return; }
		    else if (a3 == ref) { a3->flags &= (~ATOMFLAG_SELECTED); a3 = NULL; return; }
		    
		    const fGL * p1 = a1->GetCRD(0);
		    const fGL * p2 = a2->GetCRD(0);
		    const fGL * p3 = a3->GetCRD(0);
		    const fGL * p4 = ref->GetCRD(0);
		    
		    //float tor = measure_tor(p1, p2, p3, p4);		// not const-correct???
		    v3d<fGL> v1(p2, p1);
		    v3d<fGL> v2(p2, p3);
		    v3d<fGL> v3(p3, p4);
		    fGL tor = v1.tor(v2, v3) * 180.0 / M_PI;
		    
		    str1 << "torsion: " << tor << " deg " << endl << ends;
		    PrintToLog(mbuff1);
		    
		    a1->flags &= (~ATOMFLAG_SELECTED); a1 = NULL;
		    a2->flags &= (~ATOMFLAG_SELECTED); a2 = NULL;
		    a3->flags &= (~ATOMFLAG_SELECTED); a3 = NULL;
		    ref->flags &= (~ATOMFLAG_SELECTED);
		    
		    UpdateAllGraphicsViews();
		  }
	}
}

void project::DoEnergyPlot1D(i32s inda, i32s indb, i32s indc, i32s indd, i32s div1, fGL start1, fGL end1, i32s optsteps)
{
	GetCurrentSetup()->DiscardCurrentEngine();
	i32s curr_eng_index = GetCurrentSetup()->GetCurrEngIndex();
	
	engine * eng1 = GetCurrentSetup()->CreateEngineByIndex(curr_eng_index);		// with constraints.
	engine * eng2 = GetCurrentSetup()->CreateEngineByIndex(curr_eng_index);		// without constraints.
	
	i32s molnum = 0; i32s in_crdset = 0;
	
	i32s atmi1[4] = { inda, indb, indc, indd };
	atom * atmr1[4]; f64 range1[2];
	range1[0] = M_PI * start1 / 180.0;
	range1[1] = M_PI * end1 / 180.0;

	for (i32s n1 = 0;n1 < 4;n1++)
	{
		iter_al it1;
		
		it1 = FindAtomByIndex(atmi1[n1]);
		if (it1 == GetAtomsEnd()) { PrintToLog("ERROR : atom not found!\n"); return; }
		atmr1[n1] = & (* it1);
	}
	
	if (!is_groups_clean) UpdateGroups();		// for internal coordinates...
	intcrd * ic = new intcrd((* this), molnum, in_crdset);
	i32s ict1 = ic->FindTorsion(atmr1[1], atmr1[2]);
if (ict1 < 0) { PrintToLog("ERROR : could not find ic.\n"); return; }
	
	v3d<fGL> v1a(atmr1[1]->GetCRD(in_crdset), atmr1[0]->GetCRD(in_crdset));
	v3d<fGL> v1b(atmr1[1]->GetCRD(in_crdset), atmr1[2]->GetCRD(in_crdset));
	v3d<fGL> v1c(atmr1[2]->GetCRD(in_crdset), atmr1[3]->GetCRD(in_crdset));
	f64 oldt1 = v1a.tor(v1b, v1c);
	
	i32s fft1 = eng1->FindTorsion(atmr1[0], atmr1[1], atmr1[2], atmr1[3]);
if (ict1 < 0) { PrintToLog("ERROR : could not find tor-term.\n"); return; }

	plot1d_view * plot = GetGraphicsFactory()->ProducePlot1DView(this, PLOT_USERDATA_STRUCTURE, 0);
	
	f64 tor1 = range1[0];
	for (i32s s1 = 0;s1 < (div1 + 1);s1++)
	{
		ic->SetTorsion(ict1, tor1 - oldt1);
		ic->UpdateCartesian();
		
		CopyCRD(this, eng1, 0);		// lock_local_structure needs coordinates!!!
		eng1->SetTorsionConstraint(fft1, tor1, 5000.0, true);
		
		// optimize...
		
		geomopt * opt = new geomopt(eng1, 100, 0.025);		// optimal settings?!?!?
		
		for (i32s n1 = 0;n1 < optsteps;n1++)
		{
			opt->TakeCGStep(conjugate_gradient::Newton2An);
			cout << n1 << " " << opt->optval << " " << opt->optstp << endl;
			
			if (!(n1 % 50))
			{
				CopyCRD(eng1, this, 0); CenterCRDSet(0);
				UpdateAllGraphicsViews(true);
			}
		}
		
		CopyCRD(eng1, this, 0); CenterCRDSet(0);
		delete opt;
		
		// compute energy for final structure...
		
		f64 value;
		CopyCRD(this, eng2, 0);
		eng2->Compute(0);
		value = eng2->energy;
		
		// ...and add it to the plot.
		
		void * udata = convert_cset_to_plotting_udata(this, 0);
		plot->AddData(180.0 * tor1 / M_PI, value, udata);
		
		char mbuff1[256];
		strstream str1(mbuff1, sizeof(mbuff1));
		str1 << "tor = " << (180.0 * tor1 / M_PI) << " deg, energy = " << value << " kJ/mol." << endl << ends;
		PrintToLog(mbuff1);
		
		tor1 += (range1[1] - range1[0]) / (f64) div1;
	}
	
	plotting_view_vector.push_back(plot);
	plot->SetCenterAndScale();
	plot->Update();
	
	delete eng2;
	delete eng1;
	delete ic;
}

void project::DoEnergyPlot2D(i32s inda, i32s indb, i32s indc, i32s indd, i32s div1, fGL start1, fGL end1, i32s indi, i32s indj, i32s indk, i32s indl, i32s div2, fGL start2, fGL end2, i32s optsteps)
{
	GetCurrentSetup()->DiscardCurrentEngine();
	i32s curr_eng_index = GetCurrentSetup()->GetCurrEngIndex();
	
	engine * eng1 = GetCurrentSetup()->CreateEngineByIndex(curr_eng_index);		// with constraints.
	engine * eng2 = GetCurrentSetup()->CreateEngineByIndex(curr_eng_index);		// without constraints.
	
	i32s molnum = 0; i32s in_crdset = 0;
	
	i32s atmi1[4] = { inda, indb, indc, indd };
	atom * atmr1[4]; f64 range1[2];
	range1[0] = M_PI * start1 / 180.0;
	range1[1] = M_PI * end1 / 180.0;
	
	i32s atmi2[4] = { indi, indj, indk, indl };
	atom * atmr2[4]; f64 range2[2];
	range2[0] = M_PI * start2 / 180.0;
	range2[1] = M_PI * end2 / 180.0;
	
	for (i32s n1 = 0;n1 < 4;n1++)
	{
		iter_al it1;
		
		it1 = FindAtomByIndex(atmi1[n1]);
		if (it1 == GetAtomsEnd()) { PrintToLog("ERROR : tor1 atom not found!\n"); return; }
		atmr1[n1] = & (* it1);
		
		it1 = FindAtomByIndex(atmi2[n1]);
		if (it1 == GetAtomsEnd()) { PrintToLog("ERROR : tor2 atom not found!\n"); return; }
		atmr2[n1] = & (* it1);
	}
	
	if (!is_groups_clean) UpdateGroups();		// for internal coordinates...
	intcrd * ic = new intcrd((* this), molnum, in_crdset);
	i32s ict1 = ic->FindTorsion(atmr1[1], atmr1[2]);
if (ict1 < 0) { PrintToLog("ERROR : could not find ic for tor1.\n"); return; }
	i32s ict2 = ic->FindTorsion(atmr2[1], atmr2[2]);
if (ict2 < 0) { PrintToLog("ERROR : could not find ic for tor2.\n"); return; }
	
	v3d<fGL> v1a(atmr1[1]->GetCRD(in_crdset), atmr1[0]->GetCRD(in_crdset));
	v3d<fGL> v1b(atmr1[1]->GetCRD(in_crdset), atmr1[2]->GetCRD(in_crdset));
	v3d<fGL> v1c(atmr1[2]->GetCRD(in_crdset), atmr1[3]->GetCRD(in_crdset));
	f64 oldt1 = v1a.tor(v1b, v1c);
	
	v3d<fGL> v2a(atmr2[1]->GetCRD(in_crdset), atmr2[0]->GetCRD(in_crdset));
	v3d<fGL> v2b(atmr2[1]->GetCRD(in_crdset), atmr2[2]->GetCRD(in_crdset));
	v3d<fGL> v2c(atmr2[2]->GetCRD(in_crdset), atmr2[3]->GetCRD(in_crdset));
	f64 oldt2 = v2a.tor(v2b, v2c);
	
	i32s fft1 = eng1->FindTorsion(atmr1[0], atmr1[1], atmr1[2], atmr1[3]);
if (ict1 < 0) { PrintToLog("ERROR : could not find tor-term for tor1.\n"); return; }
	i32s fft2 = eng1->FindTorsion(atmr2[0], atmr2[1], atmr2[2], atmr2[3]);
if (ict2 < 0) { PrintToLog("ERROR : could not find tor-term for tor2.\n"); return; }

	plot2d_view * plot = GetGraphicsFactory()->ProducePlot2DView(this, PLOT_USERDATA_STRUCTURE, 0);
	
	f64 tor1 = range1[0];
	for (i32s s1 = 0;s1 < (div1 + 1);s1++)
	{
		f64 tor2 = range2[0];
		for (i32s s2 = 0;s2 < (div2 + 1);s2++)
		{
			ic->SetTorsion(ict1, tor1 - oldt1);
			ic->SetTorsion(ict2, tor2 - oldt2);
			ic->UpdateCartesian();
			
			CopyCRD(this, eng1, 0);		// lock_local_structure needs coordinates!!!
			eng1->SetTorsionConstraint(fft1, tor1, 5000.0, true);
			eng1->SetTorsionConstraint(fft2, tor2, 5000.0, true);
			
			// optimize...
			
			geomopt * opt = new geomopt(eng1, 100, 0.025);		// optimal settings?!?!?
			
			for (i32s n1 = 0;n1 < optsteps;n1++)
			{
				opt->TakeCGStep(conjugate_gradient::Newton2An);
				cout << n1 << " " << opt->optval << " " << opt->optstp << endl;
				
				if (!(n1 % 50))
				{
					CopyCRD(eng1, this, 0); CenterCRDSet(0);
					UpdateAllGraphicsViews(true);
				}
			}
			
			CopyCRD(eng1, this, 0); CenterCRDSet(0);
			delete opt;
			
			// compute energy for final structure...
			
			f64 value;
			CopyCRD(this, eng2, 0);
			eng2->Compute(0);
			value = eng2->energy;
			
			// ...and add it to the plot.
			
			void * udata = convert_cset_to_plotting_udata(this, 0);
			plot->AddData(180.0 * tor1 / M_PI, 180.0 * tor2 / M_PI, value, udata);
			
			char mbuff1[256];
			strstream str1(mbuff1, sizeof(mbuff1));
			str1 << "tor1 = " << (180.0 * tor1 / M_PI) << " deg, tor2 = " << (180.0 * tor2 / M_PI) << " deg, energy = " << value << " kJ/mol." << endl << ends;
			PrintToLog(mbuff1);
			
			tor2 += (range2[1] - range2[0]) / (f64) div2;
		}
		
		tor1 += (range1[1] - range1[0]) / (f64) div1;
	}
	
	plotting_view_vector.push_back(plot);
	plot->SetCenterAndScale();
	plot->Update();
	
	delete eng2;
	delete eng1;
	delete ic;
}

/*################################################################################################*/

void color_mode_element::GetColor4(const void * dd, i32s cs, fGL * pp)
{
	atom * ref = (atom *) dd;
	const fGL * color = ref->el.GetColor();
	pp[0] = color[0]; pp[1] = color[1]; pp[2] = color[2]; pp[3] = 1.0;
}

void color_mode_state::GetColor4(const void * dd, i32s cs, fGL * pp)
{
//	atom * ref = (atom *) dd;
//	mdl * mdl = ref->mdl;		// CURRENTLY BROKEN...
	
	pp[0] = 0.0; pp[1] = 0.0; pp[2] = 1.0; pp[3] = 0;	// loop
	
/*	if (mdl->chn_info == NULL) return;
	vector<chn_info *> & ci = * mdl->chn_info;
	
	if (ref->id[1] < 0 || ref->id[2] < 0) return;
	if (ci[ref->id[1]]->state == NULL) return;
	
	char state = ci[ref->id[1]]->state[ref->id[2]];
	
	switch (state)
	{
		case '4':
		p2[0] = 1.0; p2[1] = 0.0; p2[2] = 0.0;		// helix
		return;
		
		case 'S':
		p2[0] = 0.0; p2[1] = 1.0; p2[2] = 0.0;		// strand
		return;
	}	*/
}

/*################################################################################################*/

transparent_primitive::transparent_primitive(void)
{
	owner = NULL; data = NULL;
	z_distance = 0.0;
}

transparent_primitive::transparent_primitive(void * p1, transparent_primitive_data & p2)
{
	owner = p1; data = & p2;
	z_distance = 0.0;
}

transparent_primitive::transparent_primitive(const transparent_primitive & p1)
{
	owner = p1.owner; data = p1.data;
	z_distance = p1.z_distance;
}

transparent_primitive::~transparent_primitive(void)
{
}

bool transparent_primitive::TestOwner(void * p1) const
{
	return (owner == p1);
}

transparent_primitive_data * transparent_primitive::GetData(void) const
{
	return data;
}

void transparent_primitive::UpdateDistance(const fGL * crd_c, const fGL * zdir)
{
	fGL crd_p[3];
	crd_p[0] = data->midpoint[0] - crd_c[0];
	crd_p[1] = data->midpoint[1] - crd_c[1];
	crd_p[2] = data->midpoint[2] - crd_c[2];
	
	// if we mark the vector from crd_c to crd_p as v1, and angle between v1 and zdir as angle
	// alpha, we have
	//
	//	zdist = cos(alpha) * |v1|			, where |v1| = length of v1. since
	//
	//	cos(alpha) = ip(v1,zdir) / (|zdir| * |v1|)	, we have
	//
	//	zdist = ip(v1,zdir) / |zdir|			, and if zdir is a unit vector,
	//
	//	zdist = ip(v1,zdir)
	
	z_distance = crd_p[0] * zdir[0] + crd_p[1] * zdir[1] + crd_p[2] * zdir[2];
}

bool transparent_primitive::operator<(const transparent_primitive & p1) const
{
	return (z_distance > p1.z_distance);	// inverted order...
}

/*################################################################################################*/

transparent_primitive_data::transparent_primitive_data(void)
{
}

transparent_primitive_data::~transparent_primitive_data(void)
{
}

/*################################################################################################*/

tpd_tri_3c::tpd_tri_3c(fGL * c1, fGL * p1, fGL * c2, fGL * p2, fGL * c3, fGL * p3)
{
	color[0] = c1;
	color[1] = c2;
	color[2] = c3;
	
	point[0] = p1;
	point[1] = p2;
	point[2] = p3;
	
	UpdateMP();
}

tpd_tri_3c::~tpd_tri_3c(void)
{
}

void tpd_tri_3c::Render(void)
{
	glBegin(GL_TRIANGLES);
	
	glColor4fv(color[0]);
	glVertex3fv(point[0]);
	
	glColor4fv(color[1]);
	glVertex3fv(point[1]);
	
	glColor4fv(color[2]);
	glVertex3fv(point[2]);
	
	glEnd();	// GL_TRIANGLES
}

void tpd_tri_3c::UpdateMP(void)
{
	midpoint[0] = (point[0][0] + point[1][0] + point[2][0]) / 3.0;
	midpoint[1] = (point[0][1] + point[1][1] + point[2][1]) / 3.0;
	midpoint[2] = (point[0][2] + point[1][2] + point[2][2]) / 3.0;
}

/*################################################################################################*/

tpd_quad_4c::tpd_quad_4c(fGL * c1, fGL * p1, fGL * c2, fGL * p2, fGL * c3, fGL * p3, fGL * c4, fGL * p4)
{
	color[0] = c1;
	color[1] = c2;
	color[2] = c3;
	color[3] = c4;
	
	point[0] = p1;
	point[1] = p2;
	point[2] = p3;
	point[3] = p4;
	
	UpdateMP();
}

tpd_quad_4c::~tpd_quad_4c(void)
{
}

void tpd_quad_4c::Render(void)
{
	glBegin(GL_QUADS);
	
	glColor4fv(color[0]);
	glVertex3fv(point[0]);
	
	glColor4fv(color[1]);
	glVertex3fv(point[1]);
	
	glColor4fv(color[2]);
	glVertex3fv(point[2]);
	
	glColor4fv(color[3]);
	glVertex3fv(point[3]);
	
	glEnd();	// GL_QUADS
}

void tpd_quad_4c::UpdateMP(void)
{
	midpoint[0] = (point[0][0] + point[1][0] + point[2][0] + point[3][0]) / 4.0;
	midpoint[1] = (point[0][1] + point[1][1] + point[2][1] + point[3][1]) / 4.0;
	midpoint[2] = (point[0][2] + point[1][2] + point[2][2] + point[3][2]) / 4.0;
}

/*################################################################################################*/

#endif	// ENABLE_GRAPHICS

// eof
