// MOLDYN_DIALOG.CPP

// Copyright (C) 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "moldyn_dialog.h"

#include <gnome.h>

#include <iostream>
#include <strstream>
using namespace std;

/*################################################################################################*/

moldyn_dialog::moldyn_dialog(moldyn_param * p1) : glade_dialog("glade/moldyn_dialog.glade")
{
	param = p1;
	
	dialog = glade_xml_get_widget(xml, "moldyn_dialog");
	if (dialog == NULL) { cout << "moldyn_dialog : glade_xml_get_widget() failed!!!" << endl; return; }
	
	// initialize the widgets...
	
	entry_nsteps = glade_xml_get_widget(xml, "entry_nsteps");
	entry_temperature = glade_xml_get_widget(xml, "entry_temperature");
	entry_timestep = glade_xml_get_widget(xml, "entry_timestep");
	
	char buffer[64];
	
	ostrstream str_nsteps(buffer, sizeof(buffer));
	str_nsteps << param->nsteps << ends;
	gtk_entry_set_text(GTK_ENTRY(entry_nsteps), buffer);
	
	ostrstream str_grad(buffer, sizeof(buffer));
	str_grad << param->temperature << ends;
	gtk_entry_set_text(GTK_ENTRY(entry_temperature), buffer);
	
	ostrstream str_delta_e(buffer, sizeof(buffer));
	str_delta_e << param->timestep << ends;
	gtk_entry_set_text(GTK_ENTRY(entry_timestep), buffer);
	
	// connect the handlers...
	
	glade_xml_signal_connect_data(xml, "on_dialog_destroy", (GtkSignalFunc) handler_Destroy, (gpointer) this);

	glade_xml_signal_connect_data(xml, "on_button_ok_clicked", (GtkSignalFunc) handler_ButtonOK, (gpointer) this);
	glade_xml_signal_connect_data(xml, "on_button_cancel_clicked", (GtkSignalFunc) handler_ButtonCancel, (gpointer) this);
	
 	gnome_dialog_run(GNOME_DIALOG(dialog));		// MODAL
}

moldyn_dialog::~moldyn_dialog(void)
{
}

void moldyn_dialog::handler_Destroy(GtkWidget *, gpointer data)
{
	moldyn_dialog * ref = (moldyn_dialog *) data;
	cout << "handler_Destroy() : ref = " << ref << endl;
	
	delete ref;	// the object was created with new!!!
}

void moldyn_dialog::handler_ButtonOK(GtkWidget *, gpointer data)
{
	moldyn_dialog * ref = (moldyn_dialog *) data;
	cout << "handler_ButtonOK() : ref = " << ref << endl;
	
	// read the user's settings from widgets...
	
	GtkWidget * fileentry_fn = glade_xml_get_widget(ref->xml, "fileentry_fn");
	
	gchar * buffer;
	
	buffer = gtk_entry_get_text(GTK_ENTRY(ref->entry_nsteps));
	istrstream istr1(buffer, strlen(buffer) + 1);
	istr1 >> ref->param->nsteps;
	
	buffer = gtk_entry_get_text(GTK_ENTRY(ref->entry_temperature));
	istrstream istr2(buffer, strlen(buffer) + 1);
	istr2 >> ref->param->temperature;
	
	buffer = gtk_entry_get_text(GTK_ENTRY(ref->entry_timestep));
	istrstream istr3(buffer, strlen(buffer) + 1);
	istr3 >> ref->param->timestep;
	
	buffer = gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(fileentry_fn), FALSE);
	strcpy(ref->param->filename, buffer);
	
	ref->param->confirm = true;
	
	// close the dialog...
	
	gtk_widget_destroy(ref->dialog);
}

void moldyn_dialog::handler_ButtonCancel(GtkWidget *, gpointer data)
{
	moldyn_dialog * ref = (moldyn_dialog *) data;
	cout << "handler_ButtonCancel() : ref = " << ref << endl;
	
	// close the dialog...
	
	gtk_widget_destroy(ref->dialog);
}

/*################################################################################################*/

// eof
