// T3PROJECT.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "t3project.h"

#include "t3main.h"

#include "geomopt.h"
#include "moldyn.h"

#include "color.h"

#include "plane.h"
#include "surface.h"
#include "ribbon.h"

#include "file_import_dialog.h"
#include "file_export_dialog.h"

#include "setup_druid.h"
#include "geomopt_dialog.h"
#include "moldyn_dialog.h"

#include <sys/stat.h>	// for stat
#include <dirent.h>	// for scandir

#include <glade/glade.h>

#include <strstream>
using namespace std;

/*################################################################################################*/

GnomeUIInfo gnome_project::filext_submenu[] =
{
	GNOMEUIINFO_ITEM_STOCK("Import PDB...", "a special tool for importing PDB files.",
		popup_FileExtra1, GNOME_STOCK_MENU_OPEN),
	
	GNOMEUIINFO_ITEM_STOCK("Convert aa2sf", "this tool converts an all-atoms model (peptide/protein) into a simplified model.",
		popup_FileExtra2, GNOME_STOCK_MENU_OPEN),
	
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::file_submenu[] =
{
	GNOMEUIINFO_MENU_OPEN_ITEM(popup_FileOpen, NULL),

	GNOMEUIINFO_ITEM_STOCK("Import...", "import other file types...",
		popup_FileImport, GNOME_STOCK_MENU_OPEN),

	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_MENU_SAVE_AS_ITEM(popup_FileSaveAs, NULL),

	GNOMEUIINFO_ITEM_STOCK("Export...", "export other file types...",
		popup_FileExport, GNOME_STOCK_MENU_SAVE_AS),
	
//	GNOMEUIINFO_ITEM_STOCK("Export PostScript", "export graphic...",
//		popup_FileExportPS, GNOME_STOCK_MENU_SAVE_AS),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("Extra Tools", filext_submenu),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_MENU_CLOSE_ITEM(gnome_graphics_view::popup_FileClose, NULL),
	
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::selmodes_radio[] =
{
	GNOMEUIINFO_ITEM("Select Atoms", "select atoms",
		NULL, NULL),
	GNOMEUIINFO_ITEM("Select Residues", "select residues",
		NULL, NULL),
	GNOMEUIINFO_ITEM("Select Chains", "select chains",
		NULL, NULL),
	GNOMEUIINFO_ITEM("Select Molecules", "select molecules",
		NULL, NULL),

	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::selmodes_submenu[] =
{
        GNOMEUIINFO_RADIOLIST(selmodes_radio),
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::select_submenu[] =
{
	GNOMEUIINFO_ITEM("Select All", "select all atoms...",
		gnome_graphics_view::popup_SelectAll, NULL),
		
	GNOMEUIINFO_ITEM("Select None", "select nothing...",
		gnome_graphics_view::popup_SelectNone, NULL),

	GNOMEUIINFO_ITEM("Invert Selection", "invert the selection...",
		gnome_graphics_view::popup_InvertSelection, NULL),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("_Selection Mode", selmodes_submenu),
	
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::settings_submenu[] =
{
	GNOMEUIINFO_ITEM("Set Current Element...", "el...",
		gnome_mdi_app::shortcut_Element, NULL),
		
	GNOMEUIINFO_ITEM("Set Current Bondtype...", "bt...",
		gnome_mdi_app::shortcut_BondType, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::viewscreate_submenu[] =
{
	GNOMEUIINFO_ITEM("New Camera", "create a new view with a new camera...",
		gnome_graphics_view::popup_ViewsNewCam, NULL),
		
	GNOMEUIINFO_ITEM("New View", "create a new view using the current camera...",
		gnome_graphics_view::popup_ViewsNewView, NULL),
		
	GNOMEUIINFO_SEPARATOR,

	GNOMEUIINFO_ITEM("Energy-level Diagram", "create a new energy-level diagram view.",
		popup_ViewsNewEnLevDiag, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::views_submenu[] =
{
	GNOMEUIINFO_ITEM("Attach/Detach This View", "attach/detach this view...",
		gnome_graphics_view::popup_ViewsAttachDetach, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("_Create", viewscreate_submenu),
	
	GNOMEUIINFO_ITEM("Delete This View", "remove this view...",
		gnome_graphics_view::popup_ViewsDeleteView, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::lights_submenu[] =
{
	GNOMEUIINFO_ITEM("Create New Light", "create a new light object...",
		gnome_graphics_view::popup_LightsNewLight, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_TOGGLEITEM("Enable Local Lights", "loc",
		gnome_graphics_view::popup_LightsSwitchLoc, NULL),
		
	GNOMEUIINFO_TOGGLEITEM("Enable Global Lights", "glob",
		gnome_graphics_view::popup_LightsSwitchGlob, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::rendermode_style_radio[] =
{
	GNOMEUIINFO_ITEM("Ball & Stick", "bs",
		popup_RModeBallAndStick, NULL),
		
	GNOMEUIINFO_ITEM("Wireframe", "wf",
		popup_RModeWireFrame, NULL),
		
	GNOMEUIINFO_ITEM("van der Waals", "vdw",
		popup_RModeVanDerWaals, NULL),
		
	GNOMEUIINFO_ITEM("Cylinders", "cyl",
		popup_RModeCylinders, NULL),
		
	GNOMEUIINFO_ITEM("Nothing", "null",
		popup_RModeNothing, NULL),

	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::rendermode_color_radio[] =
{
	GNOMEUIINFO_ITEM("Color Elements", "color-el",
		popup_CModeElement, NULL),

	GNOMEUIINFO_ITEM("Color State", "color-st",
		popup_CModeState, NULL),

	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::rendermode_label_radio[] =
{
	GNOMEUIINFO_ITEM("Label Nothing", "lab-null",
		popup_LModeNothing, NULL),
		
	GNOMEUIINFO_ITEM("Label Index", "lab-ind",
		popup_LModeIndex, NULL),
		
	GNOMEUIINFO_ITEM("Label Charge", "lab-ch",
		popup_LModeCharge, NULL),
		
	GNOMEUIINFO_ITEM("Label Element", "lab-el",
		popup_LModeElement, NULL),
		
	GNOMEUIINFO_ITEM("Label AtomType", "lab-at",
		popup_LModeAtomType, NULL),
		
	GNOMEUIINFO_ITEM("Label BondType", "lab-bt",
		popup_LModeBondType, NULL),
	
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::rendermode_submenu[] =
{
        GNOMEUIINFO_RADIOLIST(rendermode_style_radio),
	GNOMEUIINFO_SEPARATOR,
        GNOMEUIINFO_RADIOLIST(rendermode_color_radio),
	GNOMEUIINFO_SEPARATOR,
        GNOMEUIINFO_RADIOLIST(rendermode_label_radio),
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::projection_radio[] =
{
	GNOMEUIINFO_ITEM("Perspective", "persp",
		gnome_graphics_view::popup_ProjPerspective, NULL),
		
	GNOMEUIINFO_ITEM("Orthographic", "ortho",
		gnome_graphics_view::popup_ProjOrthographic, NULL),
		
	GNOMEUIINFO_ITEM("Red-Green Stereo", "redgre",
		gnome_graphics_view::popup_ProjRedGreen, NULL),
		
	GNOMEUIINFO_ITEM("Relaxed-Eye Stereo", "releye",
		gnome_graphics_view::popup_ProjRelaxedEye, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::projection_submenu[] =
{
        GNOMEUIINFO_RADIOLIST(projection_radio),
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::render_submenu[] =
{
	GNOMEUIINFO_SUBTREE("_Views", views_submenu),
	GNOMEUIINFO_SUBTREE("_Lights", lights_submenu),
	GNOMEUIINFO_SUBTREE("_Rendering Mode", rendermode_submenu),
	GNOMEUIINFO_SUBTREE("_Projection", projection_submenu),
	
	GNOMEUIINFO_TOGGLEITEM("Quick Update Mode", "quick update...",
		gnome_graphics_view::popup_RenderQuickUpdate, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::objects_submenu[] =
{
	GNOMEUIINFO_ITEM("Ribbon", "...",
		popup_ObjRibbon, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("ESP-plane", "...",
		popup_ObjEPlane, NULL),
		
	GNOMEUIINFO_ITEM("volume-rendered ESP", "...",
		popup_ObjEVolume, NULL),
		
	GNOMEUIINFO_ITEM("ESP-surface", "...",
		popup_ObjESurface, NULL),
		
	GNOMEUIINFO_ITEM("ESP-colored vdW-surface", "...",
		popup_ObjVDWSurface, NULL),

	GNOMEUIINFO_SEPARATOR,
		
	GNOMEUIINFO_ITEM("Electron density plane", "...",
		popup_ObjEDPlane, NULL),
		
	GNOMEUIINFO_ITEM("Electron density surface", "...",
		popup_ObjEDSurface, NULL),
		
	GNOMEUIINFO_ITEM("Molecular orbital plane", "...",
		popup_ObjMOPlane, NULL),
		
	GNOMEUIINFO_ITEM("Molecular orbital volume", "...",
		popup_ObjMOVolume, NULL),
		
	GNOMEUIINFO_ITEM("Molecular orbital surface", "...",
		popup_ObjMOSurface, NULL),
		
	GNOMEUIINFO_ITEM("MO-density plane", "...",
		popup_ObjMODPlane, NULL),
		
	GNOMEUIINFO_ITEM("MO-density volume", "...",
		popup_ObjMODVolume, NULL),
		
	GNOMEUIINFO_ITEM("MO-density surface", "...",
		popup_ObjMODSurface, NULL),
		
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Delete Current Object", "delete",
		gnome_graphics_view::popup_ObjectsDeleteCurrent, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::compute_submenu[] =
{
	GNOMEUIINFO_ITEM("Setup...", "...",
		popup_CompSetup, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Energy", "...",
		popup_CompEnergy, NULL),
		
	GNOMEUIINFO_ITEM("Geometry Optimization...", "...",
		popup_CompGeomOpt, NULL),
		
	GNOMEUIINFO_ITEM("Molecular Dynamics...", "...",
		popup_CompMolDyn, NULL),
		
	GNOMEUIINFO_ITEM("Random Conformational Search...", "...",
		popup_CompRandomSearch, NULL),
	
	GNOMEUIINFO_ITEM("Systematic Conformational Search...", "...",
		popup_CompSystematicSearch, NULL),
	
	GNOMEUIINFO_ITEM("Monte Carlo Search...", "...",
		popup_CompMonteCarloSearch, NULL),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Plot Energy vs. 1 Torsion Angle...", "...",
		popup_CompEPlot1D, NULL),
	
	GNOMEUIINFO_ITEM("Plot Energy vs. 2 Torsion Angles...", "...",
		popup_CompEPlot2D, NULL),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Formula", "...",
		popup_CompFormula, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::hydrogens_submenu[] =
{
	GNOMEUIINFO_ITEM("Add", "...",
		popup_HAdd, NULL),
		
	GNOMEUIINFO_ITEM("Remove", "...",
		popup_HRemove, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::library_submenu[] =
{
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::build_submenu[] =
{
	GNOMEUIINFO_SUBTREE("_Hydrogens", hydrogens_submenu),
	
	GNOMEUIINFO_SUBTREE("Library", library_submenu),

	GNOMEUIINFO_ITEM("Sequence Builder (amino)...", "...",
		popup_BuilderAmino, NULL),
		
	GNOMEUIINFO_ITEM("Sequence Builder (nucleic)...", "...",
		popup_BuilderNucleic, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Center", "...",
		popup_Center, NULL),
		
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Zap All", "...",
		popup_Clear, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::popupmenu_info[] =
{
	GNOMEUIINFO_SUBTREE("_File", file_submenu),
	GNOMEUIINFO_SUBTREE("_Select", select_submenu),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("_Settings...", settings_submenu),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("_Render", render_submenu),
	GNOMEUIINFO_SUBTREE("_Objects", objects_submenu),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_SUBTREE("_Compute", compute_submenu),
	
	GNOMEUIINFO_ITEM("MD Trajectory Viewer...", "...",
		popup_TrajView, NULL),

	GNOMEUIINFO_ITEM("Set Current Orbital", "...",
		popup_SetOrbital, NULL),
	
	GNOMEUIINFO_SEPARATOR,

	GNOMEUIINFO_SUBTREE("_Build", build_submenu),
	
	GNOMEUIINFO_SEPARATOR,
	
	GNOMEUIINFO_ITEM("Enter a Command...", "Type in a command string and process it.",
		popup_EnterCommand, NULL),
		
	GNOMEUIINFO_END
};

GnomeUIInfo gnome_project::popupmenu_dummy[] =
{
        GNOMEUIINFO_SUBTREE("Ghemical", popupmenu_info),
        GNOMEUIINFO_END
};

/*################################################################################################*/

gnome_project::gnome_project(gnome_class_factory & p2) : project(p2)
{
	child = NULL;
	importpdb_mdata = NULL;		// temporary?!?!?!
	
  // new (with tearoff):
  GtkMenuItem *mi;
//  gnome_popup_menu_new(popupmenu_dummy);
//  mi = GTK_MENU_ITEM(popupmenu_dummy[0].widget);
//  popupmenu = GTK_WIDGET(mi->submenu);

  popupmenu = gnome_popup_menu_new(popupmenu_info);	// no tearoff...
  
  // create builder "Library" submenu
  mi = GTK_MENU_ITEM(build_submenu[1].widget);
  
  struct stat stat_buf;
  char buffer [1024];
  ostrstream str(buffer, sizeof(buffer));
  str << PRIMARY_PATH << "library" << DIR_SEPARATOR << ends;
  if ( stat(buffer, &stat_buf) == 0 && S_ISDIR(stat_buf.st_mode) )
    MakeLibraryWidgets(mi,
		     model_prefs->Find("MM1Interface/LibraryDir", buffer),		// memleak? never deleted???
		     model_prefs->Value("MM1Interface/LibraryLevels", 1));
  else
    MakeLibraryWidgets(mi,
		       model_prefs->Find("MM1Interface/LibraryDir", "library/"),	// memleak? never deleted???
		       model_prefs->Value("MM1Interface/LibraryLevels", 1));
	
}

gnome_project::~gnome_project(void)
{
	// check the graphics_view_vector for any "detached" views, and call gtk_widget_destroy for them.
	// the "attached" views, the ones that are linked to the GtkNotebook, will be destroyed with the notebook.
	// ok, this is a bit ugly solution, but see gnome_graphics_view::popup_ViewsAttachDetach() for more comments...
	
	for (i32u n1 = 0;n1 < graphics_view_vector.size();n1++)
	{
		gnome_graphics_view * gv = dynamic_cast<gnome_graphics_view *>(graphics_view_vector[n1]);
		if (gv != NULL && gv->detached != NULL) gtk_widget_destroy(GTK_WIDGET(gv->detached));
	}
}

// this is effectively a ctor for gnome_project, but it's moved here to avoid pure virtual function calls
// (GetFullProjectFileName(), at least). This must be called right after ctor returns, and only once!!!

// any SMARTER and MORE SECURE ways to do this???

void gnome_project::InitGnomeMDIChild(void)
{
	if (child != NULL)
	{
		cout << "BUG: multiple calls to gnome_project::InitGnomeMDIChild()." << endl;
		exit(EXIT_FAILURE);
	}
	
	char buffer[1024];
	GetFullProjectFileName(buffer, 1024);
	
	child = gnome_mdi_generic_child_new(buffer);
	gtk_object_set_user_data(GTK_OBJECT(child), (project *) this);
	
	gtk_signal_connect(GTK_OBJECT(child), "destroy", GTK_SIGNAL_FUNC(gnome_mdi_app::DestroyChildHandler), NULL);
	
	gnome_mdi_generic_child_set_config_func(child, GetConfigString, NULL);
	gnome_mdi_generic_child_set_view_creator(child, GnomeViewCreator, NULL);
	
//gnome_mdi_child_set_menu_template(GNOME_MDI_CHILD(child), ???);	// we use the popup-menu...

	gnome_mdi_add_child(gnome_mdi_app::GetInstance()->mdi, GNOME_MDI_CHILD(child));
	gnome_mdi_add_view(gnome_mdi_app::GetInstance()->mdi, GNOME_MDI_CHILD(child));
	
	// always add a treelist view and a graphics view by default.
	// add the treelist one first so that the graphics one remains on top...
	
	CreateTreeListView();
	AddGraphicsView();
	
	PrintToLog("A new project created.\n");
}

char * gnome_project::GetConfigString(GnomeMDIChild * child, gpointer)
{
	return g_strdup(child->name);
}

GtkWidget * gnome_project::GnomeViewCreator(GnomeMDIChild * child, gpointer)
{
	project * prj1 = (project *) gtk_object_get_user_data(GTK_OBJECT(child));
	gnome_project * prj2 = dynamic_cast<gnome_project *>(prj1);
	
	// the viewing widget is a vertically paned widget.
	// it contains the notebook and the text log.
	
	prj2->paned_widget = GTK_WIDGET(gtk_vpaned_new());
	
	prj2->notebook = gtk_notebook_new();
	
	prj2->logpos = 0;	// the position counter for gtk_editable_insert_text.
	prj2->scroll_widget = gtk_scrolled_window_new(NULL, NULL);
	//prj2->log_buffer = gtk_text_buffer_new(NULL);	in GTK2
	prj2->log_widget = gtk_text_new(NULL, NULL);		//gtk_text_view_new_with_buffer(prj2->log_buffer);	in GTK2
	gtk_container_add(GTK_CONTAINER(prj2->scroll_widget), GTK_WIDGET(prj2->log_widget));
	gtk_widget_show(GTK_WIDGET(prj2->log_widget)); gtk_widget_show(GTK_WIDGET(prj2->scroll_widget));
	
	gtk_paned_add1(GTK_PANED(prj2->paned_widget), GTK_WIDGET(prj2->notebook));
	gtk_paned_add2(GTK_PANED(prj2->paned_widget), GTK_WIDGET(prj2->scroll_widget));
	
	return prj2->paned_widget;
}

void gnome_project::PrintToLog(const char * message)
{
	gtk_editable_insert_text(GTK_EDITABLE(log_widget), message, strlen(message), & logpos);
}

void gnome_project::UpdateAllWindowTitles()
{
	// mdi_child_name and view titles are somewhat different issues now; check this!!!
	// we should change mdi_child_name when the project name changes, but this function is for setting the view titles.
	
	char buffer[1024];
	GetProjectFileName(buffer, 1024, true);
	
	gnome_mdi_child_set_name(GNOME_MDI_CHILD(child), buffer);
	project::UpdateAllWindowTitles();
	
	event_SelectedObjectChanged();		// update the statusbar too
}

void gnome_project::event_SelectedObjectChanged(void)
{
	// here we just set the new selection information for all statusbars...
	
	for (i32u n1 = 0;n1 < gnome_mdi_app::app_vector.size();n1++)
	{
		char txtbuff[256];
		ostrstream str(txtbuff, sizeof(txtbuff));
		
		str << project_filename << "." << GetProjectFileNameExtension() << " : ";
		if (selected_object) str << "selected object = " << selected_object->GetObjectName() << "." << ends;
		else str << "no selected object." << ends;
		
		GtkWidget * sb = gnome_mdi_app::app_vector[n1].statusbar;
		i32u context_id = gnome_mdi_app::app_vector[n1].context_id;
		
		gtk_statusbar_pop(GTK_STATUSBAR(sb), context_id);
		gtk_statusbar_push(GTK_STATUSBAR(sb), context_id, txtbuff);
	}
}

/*##############################################*/
/*##############################################*/

// the mm1-dependent popup-menu callbacks are here...
// the mm1-dependent popup-menu callbacks are here...
// the mm1-dependent popup-menu callbacks are here...

void gnome_project::popup_FileOpen(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new file_open_dialog(prj);		// will call delete itself...
}

void gnome_project::popup_FileImport(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new file_import_dialog(prj);		// will call delete itself...
}

void gnome_project::popup_FileExport(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new file_export_dialog(prj);		// will call delete itself...
}

void gnome_project::popup_FileExportGraphic(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new file_save_graphics_dialog(prj);	// will call delete itself...
}

void gnome_project::popup_FileSaveAs(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new file_save_dialog(prj);		// will call delete itself...
}

void gnome_project::popup_FileExtra1(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) new importpdb_dialog(prj);		// will call delete itself???
}

void gnome_project::popup_FileExtra2(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (prj->importpdb_mdata == NULL)
		{
			prj->err->Message("Please import a PDB file first;\nSee File/ExtraTools/ImportPDB...");
			return;
		}
		
		prj->aa2sf_ConvertAll(prj->importpdb_mdata);
		prj->UpdateAllGraphicsViews();
	}
}

void gnome_project::popup_ViewsNewEnLevDiag(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->AddEnLevDiagView();
	}
}

void gnome_project::popup_RModeWireFrame(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->render = RENDER_WIREFRAME; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_RModeBallAndStick(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->render = RENDER_BALL_AND_STICK; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_RModeVanDerWaals(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->render = RENDER_VAN_DER_WAALS; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_RModeCylinders(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->render = RENDER_CYLINDERS; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_RModeNothing(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->render = RENDER_NOTHING; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_CModeState(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->colormode = & project::cm_state; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_CModeElement(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->colormode = & project::cm_element; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeNothing(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_NOTHING; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeIndex(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_INDEX; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeCharge(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_CHARGE; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeElement(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_ELEMENT; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeAtomType(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_ATOMTYPE; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_LModeBondType(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gv->label = LABEL_BONDTYPE; gv->prj->UpdateAllGraphicsViews();
}

void gnome_project::popup_ObjRibbon(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->UpdateChains();
		DefineSecondaryStructure(prj);
		
		vector<chn_info> & ci_vector = (* prj->ref_civ);
		for (i32u n1 = 0;n1 < ci_vector.size();n1++)
		{
			if (ci_vector[n1].GetType() != chn_info::amino_acid) continue;
			if (ci_vector[n1].GetLength() < 3) continue;
			
			prj->AddObject(new ribbon(prj, gv->colormode, n1, 4));		// min. order is 2!!!
		}
		
		prj->UpdateAllGraphicsViews();
	}
}

// planes/surfaces/etc... are done here, in a simple and straightforward way.
// just equivalent stuff is repeated at TARGET2. how to improve this???

// an advanced dialog box would be needed here, to replace the text stuff.
// the dialog should be flexible enough, so it could be used in all models, not just here...

// does it make any sense to write DoPlane, DoSurface, etc???

void gnome_project::popup_ObjEPlane(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add plane esp rb1 138.0 1.0 50 1 0.75";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjEVolume(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add volrend esp rb2 138.0 1.0 25 0.50";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjESurface(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add surf2 esp unity red blue +35.0 -35.0 1.0 2.0 50 0 0 0.50";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjVDWSurface(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add surf1 vdws esp rb1 1.0 70.0 2.0 50 1 1 0.65";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjEDPlane(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add plane eldens rb1 0.05 0.75 50 1 0.75";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjEDSurface(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add surf1 eldens unity red 0.01 1.0 1.5 50 0 0 0.65";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMOPlane(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add plane mo rb1 0.05 0.75 50 1 0.75";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMOVolume(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add volrend mo rb2 0.025 1.5 25 0.50";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMOSurface(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add surf2 mo unity red blue +0.025 -0.025 1.0 1.5 50 0 0 0.50";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMODPlane(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add plane mod rb1 0.005 0.75 50 1 0.75";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMODVolume(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add volrend mod rb2 0.0025 1.5 25 0.35";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_ObjMODSurface(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		if (!prj->GetCurrentSetup()->GetCurrentEngine())
		{
			prj->err->Message("Please calculate energy first!");
		}
		else
		{
			static const char command[] = "add surf1 mod unity red 0.0025 1.0 1.5 50 0 0 0.65";
			new command_dialog(prj, gv, command);
		}
	}
}

void gnome_project::popup_CompSetup(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	
	new setup_druid(prj);		// will call delete itself...
}

void gnome_project::popup_CompEnergy(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) prj->DoEnergy();
}

void gnome_project::popup_CompGeomOpt(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		geomopt_param param;
		
		param.treshold_nsteps = 500;
		param.treshold_grad = 1.0e-3;
		param.treshold_delta_e = 1.0e-7;

		prj->DoGeomOpt(param);
	}
}

void gnome_project::popup_CompMolDyn(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		moldyn_param param;
		
		param.nsteps = 10000;
		param.temperature = 300.0;
		param.timestep = 0.5;
		
		prj->DoMolDyn(param);
	}
}

void gnome_project::popup_CompRandomSearch(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		static const char command[] = "random_search 100 250";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_CompSystematicSearch(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		static const char command[] = "systematic_search 6 250";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_CompMonteCarloSearch(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		static const char command[] = "montecarlo_search 10 100 250";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_CompEPlot1D(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
prj->err->Message("PLEASE NOTE!\nThe command string, which is displayed in the next dialog, is incomplete.\nYou should replace the letters A-D with atom indices that define the torsion.\n\nALSO NOTE: optsteps should be -1 for all setups other than ALLMM (constraints implemented in MM only).");
		
		static const char command[] = "make_plot1 A B C D 36 0.0 360.0 250";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_CompEPlot2D(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
prj->err->Message("PLEASE NOTE!\nThe command string, which is displayed in the next dialog, is incomplete.\nYou should replace the letters A-D and I-L with atom indices that define the torsions.\n\nALSO NOTE: optsteps should be -1 for all setups other than ALLMM (constraints implemented in MM only).");
		
		static const char command[] = "make_plot2 A B C D 36 0.0 360.0 I J K L 36 0.0 360.0 250";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_CompFormula(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj) prj->DoFormula();
}

void gnome_project::popup_TrajView(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (!prj->GetTrajectoryFile())
	{
		if (prj) new trajfile_dialog(prj);	// will call delete itself...
	}
	else prj->err->ErrorMessage("Trajectory already open?!?!?!");
}

void gnome_project::popup_SetOrbital(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->err->Message("PLEASE NOTE!\nThe command string, which is displayed in the next dialog, is incomplete.\nYou should replace the letter X with the orbital index that will become the current orbital.");
		
		static const char command[] = "set_current_orbital X";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_HAdd(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->AddHydrogens();
		prj->UpdateAllGraphicsViews();
	}
}

void gnome_project::popup_HRemove(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->RemoveHydrogens();
		prj->UpdateAllGraphicsViews();
	}
}

void gnome_project::popup_Clear(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj && 
	    prj->err->Question("Are you sure you want to clear everything?"))
	{
	  prj->Clear();
	  prj->UpdateAllGraphicsViews();
	}
}

void gnome_project::popup_Library(GtkWidget *item, gpointer data)
{
  project * prj1 = (project *) gtk_object_get_user_data(GTK_OBJECT(item));
  gnome_project * prj = dynamic_cast<gnome_project *>(prj1);

  static char buffer[256];
  g_snprintf(buffer, 256, "%s", data);
cout << "preparing to open a library file : " << buffer << endl;	// just for debugging...

  bool open = prj->IsEmpty();	// open, not insert
  
  ifstream ifile;
  ifile.open(buffer, ios::in);
  prj->ReadStream(ifile, true);
  
  // This should probably go into the prjiew class 
  // that is, the prjiew::ReadStream() should override 
  // the model::ReadStream() and the rest
  	
  if (!open)	// we append this file to the current project
    {
      if (prj->selected_object != NULL)  // if an object was selected,
	{				  // remove that selection and leave
	  prj->selected_object = NULL;	 // the newly added atoms as selection;
	  prj->event_SelectedObjectChanged();	// now the mouse tools also affect the
	}				 // newly added atoms!
    }
	
  ifile.close();
}

void gnome_project::popup_BuilderAmino(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
prj->err->Message("PLEASE NOTE!\nThe command string, which is displayed in the next dialog, is incomplete.\nYou should replace the default sequence AAA with the sequence to be built.");
		
		static const char command[] = "build_amino AAA";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_BuilderNucleic(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
prj->err->Message("PLEASE NOTE!\nThe command string, which is displayed in the next dialog, is incomplete.\nYou should replace the default sequence GGG with the sequence to be built.");
		
		static const char command[] = "build_nucleic GGG";
		new command_dialog(prj, gv, command);
	}
}

void gnome_project::popup_Center(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		prj->CenterCRDSet(0);
		
		// Which is the current Coord Set?
		// usually it's 0, but it could be any/all. the crdset-system is not yet in GUI...
		
		// 2001-06-18 TH: yes, this is not quite ready. but i guess in the end we will
		// move ALL crd-sets...
		
// sometimes this centering won't work, if the camera won't point to the origo.
// so, here we try to turn it there. ANY EFFECTS TO LIGHTS, ETC??????

	//	for (i32u n1 = 0;n1 < prj->camera_vector.size();n1++)
	//	{
			// rotation or translation???
			// translation would affect less to lights (since it's accurate)????
	//	}
		
		prj->UpdateAllGraphicsViews();
	}
}

void gnome_project::popup_EnterCommand(GtkWidget *, gpointer data)
{
	gnome_graphics_view * gv = gnome_graphics_view::GetGV((GtkWidget *) data);
	gnome_project * prj = dynamic_cast<gnome_project *>(gv->prj);
	if (prj)
	{
		new command_dialog(prj, gv, NULL);
	}
}

/*##############################################*/
/*##############################################*/

void gnome_project::GeomOptGetParam(geomopt_param & param)
{
	new geomopt_dialog(& param);		// the object will call delete itself...
	
	// the above dialog is a modal one; we will return from the function only after the dialog is closed.
	cout << "now returning from gnome_project::GeomOptGetParam()." << endl;
}

void gnome_project::MolDynGetParam(moldyn_param & param)
{
	new moldyn_dialog(& param);		// the object will call delete itself...
	
	// the above dialog is a modal one; we will return from the function only after the dialog is closed.
	cout << "now returning from gnome_project::MolDynGetParam()." << endl;
}

void gnome_project::MakeLibraryWidgets(GtkMenuItem *menu, 
					const char *directory_path, 
					int recursion)
{
  struct stat stat_buf;

  if ( stat(directory_path, &stat_buf) == 0 && S_ISDIR(stat_buf.st_mode) )
    {
      struct dirent *namelist;
      DIR *dirList;
      GtkWidget *library_item;

       if (( dirList = opendir(directory_path) ))
	 {
//	   char *filename = new char[1024];		// OLD
	   while (( namelist = readdir(dirList) ))
	     {
	       char *filename = new char[1024];		// a new string for each item???
	       
	       strncpy(filename, directory_path, 512);
	       strncat(filename, namelist->d_name, 511);
	       
	       if ( namelist->d_name[0] != '.' &&
		    stat(filename, &stat_buf) == 0 &&
		    strstr(namelist->d_name, ".gpr") != NULL)
		 {
		   library_item = gtk_menu_item_new_with_label(namelist->d_name);
		   
		   // we pass the string "filename" here as user data.
		   // "filename" must be unique for each item, and it must stay in memory.
		   
		   gtk_signal_connect(GTK_OBJECT(library_item), "activate",
				      GTK_SIGNAL_FUNC(popup_Library),
				      filename);
		   
		   gtk_object_set_user_data(GTK_OBJECT(library_item), 
					    (project *) this);

		   gtk_widget_show(library_item);

		   gtk_menu_append(GTK_MENU(menu->submenu),
				   library_item);
		 }
	       else if ( namelist->d_name[0] != '.' &&
			 stat(filename, &stat_buf) == 0 &&
			 S_ISDIR(stat_buf.st_mode) &&
			 recursion != 0)
		 {
		   library_item = gtk_menu_item_new_with_label(namelist->d_name);
		   gtk_menu_item_set_submenu(GTK_MENU_ITEM(library_item),
					     gtk_menu_new());

		   gtk_widget_show(library_item);
		   
		   gtk_menu_append(GTK_MENU(menu->submenu),
				   library_item);
		   
		   strcat(filename, "/");
		   MakeLibraryWidgets(GTK_MENU_ITEM(library_item), filename, recursion-1);
		 }
	// this would be the correct place to delete the string, if it were temporary.
	// however, we must retain the strings in memory because we will use them later, so this will create a (small) memory leak.
	// we should either store the strings to a static block of memory, or someway arrange the memory release when program stops.
//	   delete [] filename;		// can't do this, or we might use freed (and possibly overwritten) memory later...
	     }
//	   delete [] filename;		// OLD
	   closedir(dirList);
	 }

     }
}

/*################################################################################################*/

// eof
